def assert_database_state(state1, state2):
    assert state1 == state2


class HashableDict(dict):
    def __hash__(self):
        return hash((frozenset(self), frozenset(self.values())))

    def __eq__(self, other):
        return super().__eq__(other)


def to_hashable(items):
    for key, value in items:
        if key == "a":
            yield key, tuple(tuple(v) for v in value)

        elif key == "b":
            yield (
                key,
                tuple(
                    (atom1, atom2, order, tuple(periodicity))
                    for atom1, atom2, order, periodicity in value
                ),
            )

        elif key == "m":
            yield key, tuple(tuple(v) for v in value)

        elif key == "BB":
            yield key, tuple(map(HashableDict, value))

        elif key == "aI":
            yield key, tuple(tuple(v) for v in value)

        elif key == "bI":
            yield key, tuple(tuple(v) for v in value)

        elif key == "nBB":
            yield key, tuple(value)

        else:
            yield key, value


def get_entry(item):
    """
    Create a :class:`.DatabaseEntry` from a :class:`dict`.

    Parameters
    ----------
    item : :class:`dict`
        A entry retrieved from a MongoDB collection.

    Returns
    -------
    :class:`.DatabaseEntry`
        The entry created from `item`.

    """

    return DatabaseEntry(
        **{
            key: value
            for key, value in to_hashable(item.items())
            if key != "_id"
        }
    )


class DatabaseEntry:
    """
    Represents an entry in a MongoDB collection.

    """

    def __init__(self, **kwargs):
        """
        Initialize a :class:`.DatabaseEntry`.

        Parameters
        ----------
        kwargs : :class:`dict`
            The data held by the entry.

        """

        self._items = kwargs

    def __eq__(self, other):
        return self._items == other._items

    def __hash__(self):
        return hash(
            (
                frozenset(self._items),
                frozenset(self._items.values()),
            )
        )

    def __repr__(self):
        return f"DatabaseEntry({self._items})"

    def __str__(self):
        return repr(self)


class DatabaseState:
    """
    Represents the state of a MongoDB database.

    """

    def __init__(self, entries):
        """
        Initialize a :class:`.DatabaseState`.

        Parameters
        ----------
        entries : :class:`dict`
            Maps a :class:`.DatabaseEntry` instance to the number
            of times an entry with that data appears in the database.

        """

        self._entries = dict(entries)

    def __eq__(self, other):
        return self._entries == other._entries

    def __repr__(self):
        return f"DatabaseState({self._entries})"

    def __str__(self):
        return repr(self)



class CaseData:
    """
    A test case.

    Attributes
    ----------
    molecule : :class:`.Molecule`
        The molecule to test.

    path : :class:`str`
        The name of the structure file to use for the test.

    """

    def __init__(self, molecule, path):
        """
        Initialize a :class:`.CaseData` instance.

        Parameters
        ----------
        molecule : :class:`.Molecule`
            The molecule to test.

        path : :class:`str`
            The name of the structure file to use for the test.

        """

        self.molecule = molecule
        self.path = path



import typing
from collections.abc import Callable, Iterable, Iterator
from functools import partial

import numpy as np

from stk._internal.building_block import BuildingBlock
from stk._internal.ea.molecule_record import MoleculeRecord
from stk._internal.ea.mutation.record import MutationRecord
from stk._internal.key_makers.inchi import Inchi
from stk._internal.key_makers.molecule import MoleculeKeyMaker
from stk._internal.topology_graphs.topology_graph.topology_graph import (
    TopologyGraph,
)
from stk._internal.utilities.utilities import dice_similarity

T = typing.TypeVar("T", bound=TopologyGraph)


class SimilarBuildingBlock:
    """
    Substitutes similar building blocks.

    This mutator takes a :class:`.ConstructedMolecule` and substitutes
    the building blocks with the most similar one from a given set.
    Repeated mutations on the same molecule will substituted the next
    most similar molecule from the set.

    Examples:

        *Constructed Molecule Mutation*

        .. testcode:: constructed-molecule-mutation

            import stk

            # Create a molecule which is to be mutated.
            bb1 = stk.BuildingBlock('NCCN', [stk.PrimaryAminoFactory()])
            bb2 = stk.BuildingBlock('O=CCC=O', [stk.AldehydeFactory()])
            polymer = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear((bb1, bb2), 'AB', 3),
            )

            # Create molecules used to substitute building blocks.
            building_blocks = (
                stk.BuildingBlock(
                    smiles='NC[Si]CCN',
                    functional_groups=[stk.PrimaryAminoFactory()],
                ),
                stk.BuildingBlock(
                    smiles='NCCCCCCCN',
                    functional_groups=[stk.PrimaryAminoFactory()],
                ),
                stk.BuildingBlock(
                    smiles='NC1CCCCC1N',
                    functional_groups=[stk.PrimaryAminoFactory()],
                ),
            )

            # Create the mutator.

            def has_primary_amino_group(building_block):
                fg, = building_block.get_functional_groups(0)
                return type(fg) is stk.PrimaryAmino

            similar_bb = stk.SimilarBuildingBlock(
                building_blocks=building_blocks,
                is_replaceable=has_primary_amino_group,
            )

            # Mutate a molecule.
            mutation_record1 = similar_bb.mutate(polymer)

            # Mutate the molecule a second time.
            mutation_record2 = similar_bb.mutate(polymer)

    """

    def __init__(
        self,
        building_blocks: Iterable[BuildingBlock],
        is_replaceable: Callable[[BuildingBlock], bool],
        key_maker: MoleculeKeyMaker = Inchi(),
        name: str = "SimilarBuildingBlock",
        random_seed: int | np.random.Generator | None = None,
    ) -> None:
        """
        Parameters:

            building_blocks (list[BuildingBlock]):
                A group of molecules which are used to replace building
                blocks in molecules being mutated.

            is_replaceable:
                This function is applied to every building block in
                the molecule being mutated. Building blocks
                which returned ``True`` are liable for substitution
                by one of the molecules in `building_blocks`.

            key_maker:
                Molecules which return the same key, will iterate through
                the same set of similar molecules.

            name:
                A name to help identify the mutator instance.

            random_seed:
                The random seed to use.
        """
        if random_seed is None or isinstance(random_seed, int):
            random_seed = np.random.default_rng(random_seed)
        self._building_blocks = tuple(building_blocks)
        self._is_replaceable = is_replaceable
        self._key_maker = key_maker
        self._name = name
        self._generator = random_seed
        self._similar_building_blocks: dict[
            typing.Any, dict[typing.Any, Iterator[BuildingBlock]]
        ] = {}

    def mutate(
        self,
        record: MoleculeRecord[T],
    ) -> MutationRecord[MoleculeRecord[T]]:
        """
        Return a mutant of `record`.

        Parameters:
            record:
                The molecule to be mutated.

        Returns:
            A record of the mutation.
        """
        key = self._key_maker.get_key(record.get_molecule())
        if key not in self._similar_building_blocks:
            # Maps the key to a dict. The dict maps each
            # building block to an iterator.
            # The iterators yield the next most similar molecules in
            # `building_blocks` to the building block.
            self._similar_building_blocks[key] = {}

        similar_building_blocks = self._similar_building_blocks[key]

        # Choose the building block which undergoes mutation.
        replaceable_building_blocks = tuple(
            filter(
                self._is_replaceable,
                (
                    bb
                    for bb in record.get_molecule().get_building_blocks()
                    # TODO: this is actually a type error -- maybe
                    # get_building_blocks needs to return BuildingBlock
                    # instances?
                    if isinstance(bb, BuildingBlock)
                ),
            )
        )
        replaced_building_block = self._generator.choice(
            a=replaceable_building_blocks,  # type: ignore
        )

        # If the building block has not been chosen before, create an
        # iterator yielding similar molecules from `building_blocks`
        # for it.
        replaced_key = self._key_maker.get_key(replaced_building_block)
        if replaced_key not in similar_building_blocks:
            similar_building_blocks[replaced_key] = iter(
                sorted(
                    self._building_blocks,
                    key=partial(dice_similarity, replaced_building_block),
                    reverse=True,
                )
            )

        try:
            replacement = next(similar_building_blocks[replaced_key])
        except StopIteration:
            similar_building_blocks[replaced_key] = iter(
                sorted(
                    self._building_blocks,
                    key=partial(dice_similarity, replaced_building_block),
                    reverse=True,
                )
            )
            replacement = next(similar_building_blocks[replaced_key])

        # If the most similar molecule in `building_blocks` is itself,
        # then take the next most similar one.
        if self._key_maker.get_key(replacement) == replaced_key:
            try:
                replacement = next(similar_building_blocks[replaced_key])
            except StopIteration:
                similar_building_blocks[replaced_key] = iter(
                    sorted(
                        self._building_blocks,
                        key=partial(
                            dice_similarity,
                            replaced_building_block,
                        ),
                        reverse=True,
                    )
                )
                replacement = next(similar_building_blocks[replaced_key])

        # Build the new ConstructedMolecule.
        graph = record.get_topology_graph().with_building_blocks(
            {
                replaced_building_block: replacement,
            }
        )
        return MutationRecord(
            molecule_record=MoleculeRecord(graph),
            mutator_name=self._name,
        )



"""
Periodic Construction Result
============================

"""

from stk._internal.periodic_info import PeriodicInfo

from .construction_result import ConstructionResult


class PeriodicConstructionResult(ConstructionResult):
    """
    The result of :meth:`.TopologyGraph.construct` with periodic info.

    """

    __slots__ = [
        "_periodic_info",
    ]

    def __init__(
        self,
        construction_state,
        lattice_size,
    ):
        """
        Initialize a :class:`.ConstructionResult`.

        Parameters
        ----------
        construction_state : :class:`.ConstructionState`
            The state from which the result is initialized.

        lattice_size : :class:`tuple` of :class:`int`
            The size of the lattice in the x, y and z directions.

        """

        super().__init__(construction_state)
        self._periodic_info = PeriodicInfo(
            *(
                lattice_constant * dim
                for lattice_constant, dim in zip(
                    construction_state.get_lattice_constants(),
                    lattice_size,
                )
            )
        )

    def get_periodic_info(self):
        """
        Get the periodic cell information of the constructed molecule.

        Returns
        -------
        :class:`.PeriodicInfo`
            The periodic cell information of the constructed molecule.

        """

        return self._periodic_info



def test_get_atom_infos(case_data):
    """
    Test :meth:`.ConstructedMolecule.get_atom_infos`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the constructed molecule to test and the
        correct number of new atoms.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_get_atom_infos(
        constructed_molecule=case_data.constructed_molecule,
        num_new_atoms=case_data.num_new_atoms,
    )


def _test_get_atom_infos(constructed_molecule, num_new_atoms):
    """
    Test :meth:`.ConstructedMolecule.get_atom_infos`.

    Parameters
    ----------
    constructed_molecule : :class:`.ConstructedMolecule`
        The constructed molecule to test.

    num_new_atoms : :class:`int`
        The correct number of new atoms.

    Returns
    -------
    None : :class:`NoneType`

    """

    new_atoms = filter(
        lambda atom_info: atom_info.get_building_block() is None,
        constructed_molecule.get_atom_infos(),
    )
    assert sum(1 for _ in new_atoms) == num_new_atoms
    assert constructed_molecule.get_num_atoms() == sum(
        1 for _ in constructed_molecule.get_atom_infos()
    )



import pytest
import stk

from ....case_data import CaseData
from .utilities import get_closest_point, get_edges, get_fg_position


@pytest.fixture
def tail_2(position, building_block_2):
    point1, point2 = points = (
        position + [-10, 0, 0],
        position + [10, 0, 0],
    )

    def get_fg0_point(building_block):
        return get_closest_point(
            points=points,
            point=get_fg_position(0, building_block),
        )

    def get_fg1_point(building_block):
        return get_closest_point(
            points=points,
            point=get_fg_position(1, building_block),
        )

    vertex = stk.polymer.TailVertex(0, position, False)
    return CaseData(
        vertex=vertex,
        edges=(tuple(get_edges(vertex))[0],),
        building_block=building_block_2,
        position=position,
        alignment_tests={
            get_fg0_point: point1,
            get_fg1_point: point2,
        },
        functional_group_edges={0: 0},
    )



"""
Molecular Utilities
===================

This module defines utilities for molecules.

"""


def sort_bond_atoms_by_id(bond):
    if bond.get_atom1().get_id() < bond.get_atom2().get_id():
        return bond
    elif bond.get_order() == 9:
        # Bond ordering in Dative bonds cannot change.
        return bond
    else:
        return bond.__class__(
            atom1=bond.get_atom2(),
            atom2=bond.get_atom1(),
            order=bond.get_order(),
            periodicity=bond.get_periodicity(),
        )


def get_bond_atom_ids(bond):
    return sorted(
        (
            bond.get_atom1().get_id(),
            bond.get_atom2().get_id(),
        )
    )


def get_bond_info_atom_ids(bond_info):
    return get_bond_atom_ids(bond_info.get_bond())



import typing

import rdkit.Chem.AllChem as rdkit

from stk._internal.utilities.utilities import remake


def _with_structure_from_pdb(self, path: str) -> typing.Self:  # type: ignore[misc]
    """
    Change structure to match a ``.pdb`` file.

    Parameters:

        path:
            The full path of the ``.pdb`` file from which the structure
            should be updated.

    Returns:
        The molecule.

    """

    rdk_mol = rdkit.MolFromPDBFile(
        pdbFileName=path,
        sanitize=False,
        removeHs=False,
    )
    if rdk_mol is None:
        rdkit.MolFromPDBFile(
            molFileName=path,
            sanitize=False,
            removeHs=False,
        )

    molecule = remake(rdk_mol)
    return self._with_position_matrix(
        position_matrix=molecule.GetConformer().GetPositions()
    )



"""
M2L4 Lantern
============

"""

import numpy as np

from stk._internal.topology_graphs.edge import Edge

from .cage import Cage
from .vertices import LinearVertex, NonLinearVertex


class M2L4Lantern(Cage):
    """
    Represents a cage topology graph.

    Unoptimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb1 = stk.BuildingBlock(
            smiles='[Pd+2]',
            functional_groups=(
                stk.SingleAtom(stk.Pd(0, charge=2))
                for i in range(4)
            ),
            position_matrix=[[0, 0, 0]],
        )

        bb2 = stk.BuildingBlock(
            smiles=(
                'C1=NC=CC(C2=CC=CC(C3=C'
                'C=NC=C3)=C2)=C1'
            ),
            functional_groups=[
                stk.SmartsFunctionalGroupFactory(
                    smarts='[#6]~[#7X2]~[#6]',
                    bonders=(1, ),
                    deleters=(),
                ),
            ],
        )

        cage = stk.ConstructedMolecule(
            topology_graph=stk.cage.M2L4Lantern(
                building_blocks=(bb1, bb2),
            ),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cage.get_atoms(),
                    cage.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=(
                        1
                        if bond.get_order() == 9
                        else bond.get_order()
                    ),
                ) for bond in cage.get_bonds()
            ),
        )

    :class:`.MCHammer` optimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb1 = stk.BuildingBlock(
            smiles='[Pd+2]',
            functional_groups=(
                stk.SingleAtom(stk.Pd(0, charge=2))
                for i in range(4)
            ),
            position_matrix=[[0, 0, 0]],
        )

        bb2 = stk.BuildingBlock(
            smiles=(
                'C1=NC=CC(C2=CC=CC(C3=C'
                'C=NC=C3)=C2)=C1'
            ),
            functional_groups=[
                stk.SmartsFunctionalGroupFactory(
                    smarts='[#6]~[#7X2]~[#6]',
                    bonders=(1, ),
                    deleters=(),
                ),
            ],
        )

        cage = stk.ConstructedMolecule(
            topology_graph=stk.cage.M2L4Lantern(
                building_blocks=(bb1, bb2),
                optimizer=stk.MCHammer(),
            ),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cage.get_atoms(),
                    cage.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=(
                        1
                        if bond.get_order() == 9
                        else bond.get_order()
                    ),
                ) for bond in cage.get_bonds()
            ),
        )

    Metal building blocks with four functional groups are
    required for this topology.

    Ligand building blocks with two functional groups are required for
    this topology.

    When using a :class:`dict` for the `building_blocks` parameter,
    as in :ref:`cage-topology-graph-examples`:
    *Multi-Building Block Cage Construction*, a
    :class:`.BuildingBlock`, with the following number of functional
    groups, needs to be assigned to each of the following vertex ids:

        | 4-functional groups: 0 to 1
        | 2-functional groups: 2 to 5

    See :class:`.Cage` for more details and examples.

    """

    _vertex_prototypes = (
        NonLinearVertex(0, np.array([0, 0.5, 0])),
        NonLinearVertex(1, np.array([0, -0.5, 0])),
        LinearVertex(2, np.array([1, 0, 0]), False),
        LinearVertex(3, np.array([0, 0, 1]), False),
        LinearVertex(4, np.array([-1, 0, 0]), False),
        LinearVertex(5, np.array([0, 0, -1]), False),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[0], _vertex_prototypes[2]),
        Edge(1, _vertex_prototypes[0], _vertex_prototypes[3]),
        Edge(2, _vertex_prototypes[0], _vertex_prototypes[4]),
        Edge(3, _vertex_prototypes[0], _vertex_prototypes[5]),
        Edge(4, _vertex_prototypes[1], _vertex_prototypes[2]),
        Edge(5, _vertex_prototypes[1], _vertex_prototypes[3]),
        Edge(6, _vertex_prototypes[1], _vertex_prototypes[4]),
        Edge(7, _vertex_prototypes[1], _vertex_prototypes[5]),
    )

    _num_windows = 4
    _num_window_types = 1



import numpy as np
import pytest

from .case_data import CaseData


@pytest.fixture(
    params=(
        CaseData(np.array([0, 0, 0]), 2, np.array([0, 0, 0])),
        CaseData(np.array([0, 0, 0]), 0, np.array([0, 0, 0])),
        CaseData(np.array([1, -1, 0]), 0, np.array([0, 0, 0])),
        CaseData(np.array([1, -1, 2]), 2, np.array([2, -2, 4])),
        CaseData(np.array([2, 4, -5]), -3, np.array([-6, -12, 15])),
        CaseData(
            start=np.array([1, -1, 2]),
            scale=np.array([10, 2, -2]),
            target=np.array([10, -2, -4]),
        ),
    ),
)
def case_data(request):
    """
    A :class:`.CaseData` instance.

    """

    return request.param



import itertools as it
from .case_data import CaseData

import stk


def test_get_placer_ids(case_data: CaseData) -> None:
    """
    Test :meth:`.BuildingBlock.get_placer_ids`.

    Parameters:
        case_data:
            A test case. Holds the building block to test and the
            correct placer ids.

    """

    _test_get_placer_ids(
        building_block=case_data.building_block,
        placer_ids=case_data.placer_ids,
    )


def _test_get_placer_ids(
    building_block: stk.BuildingBlock,
    placer_ids: tuple[int, ...],
) -> None:
    """
    Test :meth:`.BuildingBlock.get_placer_ids`.

    Parameters:
        building_block:
            The building block to test.

        placer_ids:
            The correct placer ids.

    """

    for placer1, placer2 in it.zip_longest(
        building_block.get_placer_ids(),
        placer_ids,
    ):
        assert placer1 == placer2



import numpy as np
import pytest
from pytest_lazyfixture import lazy_fixture

# Fixtures must be visible for lazy_fixture() calls.
from .fixtures import *  # noqa


@pytest.fixture(
    params=(
        lazy_fixture("cof1"),
        lazy_fixture("cof2"),
        lazy_fixture("cof3"),
    ),
)
def case_data(request):
    return request.param


@pytest.fixture(params=(0,))
def aligner_edge(request):
    return request.param


@pytest.fixture(params=(0,))
def id(request):
    return request.param


@pytest.fixture(
    params=(
        np.array([0, 0, 0]),
        np.array([-20, 1, 21]),
    ),
)
def cell(request):
    return np.array(request.param)


@pytest.fixture(
    params=(
        [0, 0, 0],
        [1, 2, -20],
    ),
)
def position(request):
    return np.array(request.param, dtype=np.float64)



from collections import abc

import numpy as np

from stk._internal.building_block import BuildingBlock
from stk._internal.optimizers.null import NullOptimizer
from stk._internal.optimizers.optimizer import Optimizer
from stk._internal.reaction_factories.generic_reaction_factory import (
    GenericReactionFactory,
)
from stk._internal.reaction_factories.reaction_factory import ReactionFactory
from stk._internal.topology_graphs.edge import Edge

from .cof import Cof
from .vertices import LinearVertex, NonLinearVertex


class PeriodicHexagonal(Cof):
    """
    Represents a periodic hexagonal COF topology graph.

    Unoptimzed construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb1 = stk.BuildingBlock('BrCCBr', [stk.BromoFactory()])
        bb2 = stk.BuildingBlock(
            smiles='Brc1c(Br)c(Br)c(Br)c(Br)c1Br',
            functional_groups=[stk.BromoFactory()],
        )

        cof = stk.ConstructedMolecule(
            topology_graph=stk.cof.PeriodicHexagonal(
                building_blocks=(bb1, bb2),
                lattice_size=(2, 2, 1),
            ),
        )
        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cof.get_atoms(),
                    cof.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=bond.get_order(),
                ) for bond in cof.get_bonds()
                if all(p == 0 for p in bond.get_periodicity())
            ),
        )

    ``Collapser(scale_steps=False)`` optimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb1 = stk.BuildingBlock('BrCCBr', [stk.BromoFactory()])
        bb2 = stk.BuildingBlock(
            smiles='Brc1c(Br)c(Br)c(Br)c(Br)c1Br',
            functional_groups=[stk.BromoFactory()],
        )

        cof = stk.ConstructedMolecule(
            topology_graph=stk.cof.PeriodicHexagonal(
                building_blocks=(bb1, bb2),
                lattice_size=(2, 2, 1),
                optimizer=stk.Collapser(scale_steps=False),
            ),
        )
        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cof.get_atoms(),
                    cof.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=bond.get_order(),
                ) for bond in cof.get_bonds()
                if all(p == 0 for p in bond.get_periodicity())
            ),
        )

    Building blocks with six and two functional groups are required
    for this topology graph.

    When using a :class:`dict` for the `building_blocks` parameter,
    as in :ref:`cof-topology-graph-examples`:
    *Multi-Building Block COF Construction*, a
    :class:`.BuildingBlock`, with the following number of functional
    groups, needs to be assigned to each of the following vertex ids:

        | 6-functional groups: 0 to 3
        | 2-functional groups: 4 to 15

    Note that optimizers may not optimize the :class:`.PeriodicInfo`.
    The documentation of the optimizer will state if it does.

    See :class:`.Cof` for more details and examples.

    """

    def __init__(
        self,
        building_blocks: (
            abc.Iterable[BuildingBlock] | dict[BuildingBlock, tuple[int, ...]]
        ),
        lattice_size: tuple[int, int, int],
        vertex_alignments: dict[int, int] | None = None,
        reaction_factory: ReactionFactory = GenericReactionFactory(),
        num_processes: int = 1,
        optimizer: Optimizer = NullOptimizer(),
        scale_multiplier: float = 1.0,
    ) -> None:
        """
        Initialize a :class:`.PeriodicHexagonal` instance.

        Parameters:

            building_blocks:
                Can be a :class:`tuple` of :class:`.BuildingBlock`
                instances, which should be placed on the topology graph.

                Can also be a :class:`dict` which maps the
                :class:`.BuildingBlock` instances to the ids of the
                vertices it should be placed on. A :class:`dict` is
                required when there are multiple building blocks with the
                same number of functional groups, because in this case
                the desired placement is ambiguous.

            lattice_size:
                The size of the lattice in the x, y and z directions.

            vertex_alignments:
                A mapping from the id of a :class:`.Vertex`
                to an :class:`.Edge` connected to it.
                The :class:`.Edge` is used to align the first
                :class:`.FunctionalGroup` of a :class:`.BuildingBlock`
                placed on that vertex. Only vertices which need to have
                their default edge changed need to be present in the
                :class:`dict`. If ``None`` then the default edge is used
                for each vertex. Changing which :class:`.Edge` is used will
                mean that the topology graph represents different
                structural isomers. The edge is referred to by a number
                between ``0`` (inclusive) and the number of edges the
                vertex is connected to (exclusive).

            reaction_factory:
                The reaction factory to use for creating bonds between
                building blocks.

            num_processes:
                The number of parallel processes to create during
                :meth:`construct`.

            optimizer:
                Used to optimize the structure of the constructed
                molecule.

            scale_multiplier:
                Scales the positions of the vertices.

        Raises:

            :class:`AssertionError`
                If the any building block does not have a
                valid number of functional groups.

            :class:`ValueError`
                If the there are multiple building blocks with the
                same number of functional_groups in `building_blocks`,
                and they are not explicitly assigned to vertices. The
                desired placement of building blocks is ambiguous in
                this case.

            :class:`~.cof.UnoccupiedVertexError`
                If a vertex of the COF topology graph does not have a
                building block placed on it.

            :class:`~.cof.OverlyOccupiedVertexError`
                If a vertex of the COF topology graph has more than one
                building block placed on it.

        """

        super().__init__(
            building_blocks=building_blocks,
            lattice_size=lattice_size,
            periodic=True,
            vertex_alignments=vertex_alignments,
            reaction_factory=reaction_factory,
            num_processes=num_processes,
            optimizer=optimizer,
            scale_multiplier=scale_multiplier,
        )

    _lattice_constants = _a, _b, _c = (
        np.array([1.0, 0.0, 0.0]),
        np.array([0.5, 0.866, 0]),
        np.array([0, 0, 5 / 1.7321]),
    )

    _non_linears = (
        NonLinearVertex(0, (1 / 4) * _a + (1 / 4) * _b + (1 / 2) * _c),
        NonLinearVertex(1, (1 / 4) * _a + (3 / 4) * _b + (1 / 2) * _c),
        NonLinearVertex(2, (3 / 4) * _a + (1 / 4) * _b + (1 / 2) * _c),
        NonLinearVertex(3, (3 / 4) * _a + (3 / 4) * _b + (1 / 2) * _c),
    )

    _vertex_prototypes = (
        *_non_linears,
        LinearVertex.init_at_center(
            id=4,
            vertices=(_non_linears[0], _non_linears[1]),
        ),
        LinearVertex.init_at_center(
            id=5,
            vertices=(_non_linears[0], _non_linears[2]),
        ),
        LinearVertex.init_at_center(
            id=6,
            vertices=(_non_linears[1], _non_linears[2]),
        ),
        LinearVertex.init_at_center(
            id=7,
            vertices=(_non_linears[1], _non_linears[3]),
        ),
        LinearVertex.init_at_center(
            id=8,
            vertices=(_non_linears[2], _non_linears[3]),
        ),
        LinearVertex.init_at_shifted_center(
            id=9,
            vertices=(_non_linears[0], _non_linears[2]),
            cell_shifts=((0, 0, 0), (-1, 0, 0)),
            lattice_constants=_lattice_constants,
        ),
        LinearVertex.init_at_shifted_center(
            id=10,
            vertices=(_non_linears[0], _non_linears[1]),
            cell_shifts=((0, 0, 0), (0, -1, 0)),
            lattice_constants=_lattice_constants,
        ),
        LinearVertex.init_at_shifted_center(
            id=11,
            vertices=(_non_linears[0], _non_linears[3]),
            cell_shifts=((0, 0, 0), (0, -1, 0)),
            lattice_constants=_lattice_constants,
        ),
        LinearVertex.init_at_shifted_center(
            id=12,
            vertices=(_non_linears[2], _non_linears[1]),
            cell_shifts=((0, 0, 0), (1, -1, 0)),
            lattice_constants=_lattice_constants,
        ),
        LinearVertex.init_at_shifted_center(
            id=13,
            vertices=(_non_linears[2], _non_linears[3]),
            cell_shifts=((0, 0, 0), (0, -1, 0)),
            lattice_constants=_lattice_constants,
        ),
        LinearVertex.init_at_shifted_center(
            id=14,
            vertices=(_non_linears[1], _non_linears[3]),
            cell_shifts=((0, 0, 0), (-1, 0, 0)),
            lattice_constants=_lattice_constants,
        ),
        LinearVertex.init_at_shifted_center(
            id=15,
            vertices=(_non_linears[3], _non_linears[0]),
            cell_shifts=((0, 0, 0), (1, 0, 0)),
            lattice_constants=_lattice_constants,
        ),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[4], _vertex_prototypes[0]),
        Edge(1, _vertex_prototypes[4], _vertex_prototypes[1]),
        Edge(2, _vertex_prototypes[5], _vertex_prototypes[0]),
        Edge(3, _vertex_prototypes[5], _vertex_prototypes[2]),
        Edge(4, _vertex_prototypes[6], _vertex_prototypes[1]),
        Edge(5, _vertex_prototypes[6], _vertex_prototypes[2]),
        Edge(6, _vertex_prototypes[7], _vertex_prototypes[1]),
        Edge(7, _vertex_prototypes[7], _vertex_prototypes[3]),
        Edge(8, _vertex_prototypes[8], _vertex_prototypes[2]),
        Edge(9, _vertex_prototypes[8], _vertex_prototypes[3]),
        Edge(10, _vertex_prototypes[9], _vertex_prototypes[0]),
        Edge(
            id=11,
            vertex1=_vertex_prototypes[9],
            vertex2=_vertex_prototypes[2],
            periodicity=(-1, 0, 0),
        ),
        Edge(12, _vertex_prototypes[10], _vertex_prototypes[0]),
        Edge(
            id=13,
            vertex1=_vertex_prototypes[10],
            vertex2=_vertex_prototypes[1],
            periodicity=(0, -1, 0),
        ),
        Edge(14, _vertex_prototypes[11], _vertex_prototypes[0]),
        Edge(
            id=15,
            vertex1=_vertex_prototypes[11],
            vertex2=_vertex_prototypes[3],
            periodicity=(0, -1, 0),
        ),
        Edge(16, _vertex_prototypes[12], _vertex_prototypes[2]),
        Edge(
            id=17,
            vertex1=_vertex_prototypes[12],
            vertex2=_vertex_prototypes[1],
            periodicity=(1, -1, 0),
        ),
        Edge(18, _vertex_prototypes[13], _vertex_prototypes[2]),
        Edge(
            id=19,
            vertex1=_vertex_prototypes[13],
            vertex2=_vertex_prototypes[3],
            periodicity=(0, -1, 0),
        ),
        Edge(20, _vertex_prototypes[14], _vertex_prototypes[1]),
        Edge(
            id=21,
            vertex1=_vertex_prototypes[14],
            vertex2=_vertex_prototypes[3],
            periodicity=(-1, 0, 0),
        ),
        Edge(22, _vertex_prototypes[15], _vertex_prototypes[3]),
        Edge(
            id=23,
            vertex1=_vertex_prototypes[15],
            vertex2=_vertex_prototypes[0],
            periodicity=(1, 0, 0),
        ),
    )



"""
PDB Writer
==========

"""

from __future__ import annotations

import pathlib
import typing

from stk._internal.molecule import Molecule
from stk._internal.periodic_info import PeriodicInfo
from stk._internal.utilities.utilities import OneOrMany


class PdbWriter:
    """
    A writer class for ``.pdb`` files.

    Examples:

        *Writing to a File with a Unit Cell*

        This writer can write to a file with the unit
        cell included for periodic molecules. Note that this always
        assumes P1 space group.

        .. testcode:: writing-to-a-file-with-a-unit-cell

            import stk

            bb1 = stk.BuildingBlock('BrCCBr', [stk.BromoFactory()])
            bb2 = stk.BuildingBlock(
                smiles='BrCC(CBr)CBr',
                functional_groups=[stk.BromoFactory()],
            )
            topology_graph = stk.cof.PeriodicHoneycomb(
                building_blocks=(bb1, bb2),
                lattice_size=(3, 3, 1),
            )
            construction_result = topology_graph.construct()
            cof = (
                stk.ConstructedMolecule.init_from_construction_result(
                    construction_result=construction_result,
                )
            )
            writer = stk.PdbWriter()
            writer.write(
                molecule=cof,
                path='cof.pdb',
                periodic_info=construction_result.get_periodic_info(),
            )

        .. testcode:: writing-to-a-file-with-a-unit-cell
            :hide:

            import os

            assert os.path.exists('cof.pdb')

        .. testcleanup:: writing-to-a-file-with-a-unit-cell

            os.remove('cof.pdb')

    """

    def _write_content(
        self,
        molecule: Molecule,
        atom_ids: typing.Optional[OneOrMany[int]],
        periodic_info: typing.Optional[PeriodicInfo] = None,
    ) -> list[str]:
        if atom_ids is None:
            atom_ids = range(molecule.get_num_atoms())
        elif isinstance(atom_ids, int):
            atom_ids = (atom_ids,)

        content = []
        if periodic_info is not None:
            # Input unit cell information.
            a = periodic_info.get_a()
            b = periodic_info.get_b()
            c = periodic_info.get_c()
            alpha = periodic_info.get_alpha()
            beta = periodic_info.get_beta()
            gamma = periodic_info.get_gamma()
            content.append(
                f"CRYST1 {a:>8.3f} {b:>8.3f} {c:>8.3f}"
                f" {alpha:>6.2f} {beta:>6.2f} {gamma:>6.2f} "
                f"P 1\n"
            )

        atom_counts: dict[str, int] = {}
        hetatm = "HETATM"
        alt_loc = ""
        res_name = "UNL"
        chain_id = ""
        res_seq = "1"
        i_code = ""
        occupancy = "1.00"
        temp_factor = "0.00"

        coords = molecule.get_position_matrix()
        # This set will be used by bonds.
        atoms = set()
        for atom_id in atom_ids:
            atoms.add(atom_id)
            (atom,) = molecule.get_atoms(atom_ids=atom_id)
            serial = atom_id + 1
            element = atom.__class__.__name__
            charge = atom.get_charge()
            atom_counts[element] = atom_counts.get(element, 0) + 1
            name = f"{element}{atom_counts[element]}"
            # Make sure the coords are no more than 8 columns wide
            # each.
            x, y, z = (i for i in coords[atom_id])

            content.append(
                f"{hetatm:<6}{serial:>5} {name:<4}"
                f"{alt_loc:<1}{res_name:<3} {chain_id:<1}"
                f"{res_seq:>4}{i_code:<1}   "
                f" {x:>7.3f} {y:>7.3f} {z:>7.3f}"
                f"{occupancy:>6}{temp_factor:>6}          "
                f"{element:>2}{charge:>2}\n"
            )

        conect = "CONECT"
        for bond in molecule.get_bonds():
            a1 = bond.get_atom1().get_id()
            a2 = bond.get_atom2().get_id()
            if a1 in atoms and a2 in atoms:
                content.append(
                    f"{conect:<6}{a1 + 1:>5}{a2 + 1:>5}               \n"
                )

        content.append("END\n")

        return content

    def to_string(
        self,
        molecule: Molecule,
        atom_ids: typing.Optional[OneOrMany[int]] = None,
        periodic_info: typing.Optional[PeriodicInfo] = None,
    ) -> str:
        """
        Get a ``.pdb`` file format string of `molecule`.

        Parameters:

            molecule:
                Molecule to write to ``.pdb`` format.

            atom_ids:
                The atom ids of atoms to write. Can be a single
                :class:`int`, if a single atom is to be used, or
                ``None``, if all atoms are to be used.

            periodic_info:
                Information about the periodic cell.

        Returns:

            A string holding the content of a ``.pdf`` file.

        """

        content = self._write_content(
            molecule=molecule,
            atom_ids=atom_ids,
            periodic_info=periodic_info,
        )

        return "".join(content)

    def write(
        self,
        molecule: Molecule,
        path: pathlib.Path | str,
        atom_ids: typing.Optional[OneOrMany[int]] = None,
        periodic_info: typing.Optional[PeriodicInfo] = None,
    ) -> None:
        """
        Write `molecule` to ``.pdb`` file format.

        Parameters:

            molecule:
                Molecule to write to ``.pdb`` format.

            path:
                The full path to the file being written.

            atom_ids:
                The atom ids of atoms to write. Can be a single
                :class:`int`, if a single atom is to be used, or
                ``None``, if all atoms are to be used.

            periodic_info:
                Information about the periodic cell.

        """

        content = self._write_content(
            molecule=molecule,
            atom_ids=atom_ids,
            periodic_info=periodic_info,
        )

        with open(path, "w") as f:
            f.write("".join(content))



import stk

from dataclasses import dataclass


@dataclass(slots=True, frozen=True)
class CaseData:
    molecule: stk.Molecule
    writer: stk.TurbomoleWriter
    string: str
    periodic_info: stk.PeriodicInfo | None



import stk


def test_put_caching(mongo_client):
    collection = "_test_put_caching"
    database_name = "_test_put_caching"
    mongo_client.drop_database(database_name)

    database = stk.ValueMongoDb(
        mongo_client=mongo_client,
        collection=collection,
        database=database_name,
    )
    molecule = stk.BuildingBlock("CCC")
    database.put(molecule, 43)
    database.put(molecule, 43)

    cache_info = database._put.cache_info()
    assert cache_info.hits == 1
    assert cache_info.misses == 1

    database.put(molecule, 40)
    cache_info = database._put.cache_info()
    assert cache_info.hits == 1
    assert cache_info.misses == 2


def test_get_caching(mongo_client):
    collection = "_test_get_caching"
    database_name = "_test_get_caching"
    mongo_client.drop_database(database_name)

    database = stk.ValueMongoDb(
        mongo_client=mongo_client,
        collection=collection,
        database=database_name,
    )
    molecule = stk.BuildingBlock("CCC")
    database.put(molecule, 43)
    database.get(molecule)
    database.get(molecule)

    cache_info = database._get.cache_info()
    assert cache_info.hits == 1
    assert cache_info.misses == 1



"""
M4L6 Tetrahedron with Spacer
============================

"""

import numpy as np

from stk._internal.topology_graphs.edge import Edge

from .cage import Cage
from .vertices import LinearVertex, NonLinearVertex


class M4L6TetrahedronSpacer(Cage):
    """
    Represents a cage topology graph.

    Unoptimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        iron_atom = stk.BuildingBlock(
            smiles='[Fe+2]',
            functional_groups=(
                stk.SingleAtom(stk.Fe(0, charge=2))
                for i in range(6)
            ),
            position_matrix=[[0, 0, 0]],
        )

        bb2 = stk.BuildingBlock(
            smiles='C1=NC(C=NBr)=CC=C1',
            functional_groups=[
                stk.SmartsFunctionalGroupFactory(
                    smarts='[#6]~[#7X2]~[#35]',
                    bonders=(1, ),
                    deleters=(),
                ),
                stk.SmartsFunctionalGroupFactory(
                    smarts='[#6]~[#7X2]~[#6]',
                    bonders=(1, ),
                    deleters=(),
                ),
            ],
        )

        complex = stk.ConstructedMolecule(
            topology_graph=stk.metal_complex.OctahedralDelta(
                metals=iron_atom,
                ligands=bb2,
                optimizer=stk.MCHammer(),
            ),
        )

        # Assign Bromo functional groups to the metal complex.
        iron_oct_delta = stk.BuildingBlock.init_from_molecule(
            molecule=complex,
            functional_groups=[stk.BromoFactory()],
        )

        # Define building blocks.
        bb3 = stk.BuildingBlock(
            smiles=(
                'C1=CC(=CC=C1C2=CC=C(C=C2)Br)Br'
            ),
            functional_groups=[stk.BromoFactory()],
        )

        cage = stk.ConstructedMolecule(
            topology_graph=stk.cage.M4L6TetrahedronSpacer(
                building_blocks={
                    iron_oct_delta: (0, 1, 2, 3),
                    bb3: (4, 5, 6, 7, 8, 9),
                },
            ),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cage.get_atoms(),
                    cage.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=(
                        1
                        if bond.get_order() == 9
                        else bond.get_order()
                    ),
                ) for bond in cage.get_bonds()
            ),
        )

    :class:`.MCHammer` optimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        iron_atom = stk.BuildingBlock(
            smiles='[Fe+2]',
            functional_groups=(
                stk.SingleAtom(stk.Fe(0, charge=2))
                for i in range(6)
            ),
            position_matrix=[[0, 0, 0]],
        )

        bb2 = stk.BuildingBlock(
            smiles='C1=NC(C=NBr)=CC=C1',
            functional_groups=[
                stk.SmartsFunctionalGroupFactory(
                    smarts='[#6]~[#7X2]~[#35]',
                    bonders=(1, ),
                    deleters=(),
                ),
                stk.SmartsFunctionalGroupFactory(
                    smarts='[#6]~[#7X2]~[#6]',
                    bonders=(1, ),
                    deleters=(),
                ),
            ],
        )

        complex = stk.ConstructedMolecule(
            topology_graph=stk.metal_complex.OctahedralDelta(
                metals=iron_atom,
                ligands=bb2,
                optimizer=stk.MCHammer(),
            ),
        )

        # Assign Bromo functional groups to the metal complex.
        iron_oct_delta = stk.BuildingBlock.init_from_molecule(
            molecule=complex,
            functional_groups=[stk.BromoFactory()],
        )

        # Define building blocks.
        bb3 = stk.BuildingBlock(
            smiles=(
                'C1=CC(=CC=C1C2=CC=C(C=C2)Br)Br'
            ),
            functional_groups=[stk.BromoFactory()],
        )

        cage = stk.ConstructedMolecule(
            topology_graph=stk.cage.M4L6TetrahedronSpacer(
                building_blocks={
                    iron_oct_delta: (0, 1, 2, 3),
                    bb3: (4, 5, 6, 7, 8, 9),
                },
                optimizer=stk.MCHammer(),
            ),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cage.get_atoms(),
                    cage.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=(
                        1
                        if bond.get_order() == 9
                        else bond.get_order()
                    ),
                ) for bond in cage.get_bonds()
            ),
        )

    This topology places a ditopic spacer between the vertices of the
    tetrahedron.

    When using a :class:`dict` for the `building_blocks` parameter,
    as in :ref:`cage-topology-graph-examples`:
    *Multi-Building Block Cage Construction*, a
    :class:`.BuildingBlock`, with the following number of functional
    groups, needs to be assigned to each of the following vertex ids:

        | 3-functional groups: 0 to 3
        | 2-functional groups: 4 to 9

    Examples
    --------
    *Building Metal-Organic Tetrahedron*

    Many metal-organic cages are built using a process called
    subcomponent self-assembly, which is a complex chemical process
    that occurs in solution. In :class:`.Cage`, an example is
    provided of an alchemical approach using
    :class:`.M4L6TetrahedronSpacer` to
    build these types of cages. It is alchemical because the bonds
    formed during construction are not the same as the experimental
    reaction. Instead of forming bonds at the metal centre, we create
    bonds between disconnected ligands.

    The :class:`.M4L6TetrahedronSpacer` topology is provided for cases
    where the linkers cannot be disconnected symmetrically. However,
    in the case that the linker can be disconnected
    in a symmetrical fashion, the
    :class:`.M4L6Tetrahedron` topology can be used.

    """

    _non_linears = (
        NonLinearVertex(0, np.array([0, 0, np.sqrt(6) / 2])),
        NonLinearVertex(1, np.array([-1, -np.sqrt(3) / 3, -np.sqrt(6) / 6])),
        NonLinearVertex(2, np.array([1, -np.sqrt(3) / 3, -np.sqrt(6) / 6])),
        NonLinearVertex(3, np.array([0, 2 * np.sqrt(3) / 3, -np.sqrt(6) / 6])),
    )

    _vertex_prototypes = (
        *_non_linears,
        LinearVertex.init_at_center(
            id=4,
            vertices=(_non_linears[0], _non_linears[1]),
        ),
        LinearVertex.init_at_center(
            id=5,
            vertices=(_non_linears[0], _non_linears[2]),
        ),
        LinearVertex.init_at_center(
            id=6,
            vertices=(_non_linears[0], _non_linears[3]),
        ),
        LinearVertex.init_at_center(
            id=7,
            vertices=(_non_linears[1], _non_linears[2]),
        ),
        LinearVertex.init_at_center(
            id=8,
            vertices=(_non_linears[1], _non_linears[3]),
        ),
        LinearVertex.init_at_center(
            id=9,
            vertices=(_non_linears[2], _non_linears[3]),
        ),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[0], _vertex_prototypes[4]),
        Edge(1, _vertex_prototypes[0], _vertex_prototypes[5]),
        Edge(2, _vertex_prototypes[0], _vertex_prototypes[6]),
        Edge(3, _vertex_prototypes[1], _vertex_prototypes[4]),
        Edge(4, _vertex_prototypes[1], _vertex_prototypes[7]),
        Edge(5, _vertex_prototypes[1], _vertex_prototypes[8]),
        Edge(6, _vertex_prototypes[2], _vertex_prototypes[5]),
        Edge(7, _vertex_prototypes[2], _vertex_prototypes[7]),
        Edge(8, _vertex_prototypes[2], _vertex_prototypes[9]),
        Edge(9, _vertex_prototypes[3], _vertex_prototypes[6]),
        Edge(10, _vertex_prototypes[3], _vertex_prototypes[8]),
        Edge(11, _vertex_prototypes[3], _vertex_prototypes[9]),
    )

    _num_windows = 4
    _num_window_types = 1



class CaseData:
    """
    A test case.

    Attributes
    ----------
    constructed_molecule : :class:`.ConstructedMolecule`
        The molecule to test.

    num_new_atoms : :class:`int`
        The number of atoms added by the construction process, which
        do not belong to any building block.

    num_new_bonds : :class:`int`
        The number of bonds added by the construction process, which
        do not belong to any building block.

    num_building_blocks : :class:`dict`
        Maps each building block to the number of times it should
        be present in the constructed molecule.

    building_blocks : :class:`tuple` of :class:`.BuildingBlock`
        The building blocks of the constructed molecule.

    Notes
    -----
    You may wonder why only the number of new atoms and bonds is tested
    and not which specific atoms and bonds are new. This is because the
    atomic and bond ordering of the final constructed molecule is an
    implementation detail, and therefore it is extremely hard specify
    such a test.

    """

    def __init__(
        self,
        constructed_molecule,
        num_new_atoms,
        num_new_bonds,
        num_building_blocks,
        building_blocks,
    ):
        """
        Initialize a :class:`.CaseData` instance.

        Parameters
        ----------
        constructed_molecule : :class:`.ConstructedMolecule`
            The molecule to test.

        num_new_atoms : :class:`int`
            The number of atoms added by the construction process,
            which do not belong to any building block.

        num_new_bonds : :class:`int`
            The number of bonds added by the construction process,
            which do not belong to any building block.

        num_building_blocks : :class:`dict`
            Maps each building block to the number of times it should
            be present in the constructed molecule.

        building_blocks : :class:`tuple` of :class:`.BuildingBlock`
            The building blocks of the constructed molecule.

        """

        self.constructed_molecule = constructed_molecule
        self.num_new_atoms = num_new_atoms
        self.num_new_bonds = num_new_bonds
        self.num_building_blocks = num_building_blocks
        self.building_blocks = building_blocks



import pytest

import stk

from ....case_data import CaseData
from ...building_blocks import get_pd_atom


def _get_bi_1() -> stk.BuildingBlock:
    return stk.BuildingBlock(
        smiles="NCCN",
        functional_groups=[
            stk.SmartsFunctionalGroupFactory(
                smarts="[#7]~[#6]",
                bonders=(0,),
                deleters=(),
            ),
        ],
    )


@pytest.fixture(
    scope="session",
    params=(
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                stk.metal_complex.BidentateSquarePlanar(
                    metals={get_pd_atom(): 0},
                    ligands={_get_bi_1(): (0, 1)},
                    reaction_factory=stk.DativeReactionFactory(
                        stk.GenericReactionFactory(
                            bond_orders={
                                frozenset(
                                    {
                                        stk.GenericFunctionalGroup,
                                        stk.SingleAtom,
                                    }
                                ): 9
                            }
                        )
                    ),
                )
            ),
            smiles=(
                "[H]C1([H])C([H])([H])[N]([H])([H])->[Pd+2]2(<-[N]1([H])"
                "[H])<-[N]([H])([H])C([H])([H])C([H])([H])[N]->2([H])[H]"
            ),
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                stk.metal_complex.BidentateSquarePlanar(
                    metals=get_pd_atom(),
                    ligands=_get_bi_1(),
                    reaction_factory=stk.DativeReactionFactory(
                        stk.GenericReactionFactory(
                            bond_orders={
                                frozenset(
                                    {
                                        stk.GenericFunctionalGroup,
                                        stk.SingleAtom,
                                    }
                                ): 9
                            }
                        )
                    ),
                )
            ),
            smiles=(
                "[H]C1([H])C([H])([H])[N]([H])([H])->[Pd+2]2(<-[N]1([H])"
                "[H])<-[N]([H])([H])C([H])([H])C([H])([H])[N]->2([H])[H]"
            ),
            name=name,
        ),
    ),
)
def metal_complex_bidentate_square_planar(request) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )



import pytest
import stk

from ..case_data import GenericCaseData


@pytest.fixture
def alkene(get_atom_ids):
    a, b, c, d, e, f = get_atom_ids(6)
    return _alkene(
        carbon1=stk.C(a),
        atom1=stk.C(b),
        atom2=stk.C(c),
        carbon2=stk.C(d),
        atom3=stk.C(e),
        atom4=stk.C(f),
    )


def _alkene(carbon1, atom1, atom2, carbon2, atom3, atom4):
    bonders = (carbon2,)
    deleters = (carbon1, atom1, atom2)
    return GenericCaseData(
        functional_group=stk.Alkene(
            carbon1=carbon1,
            atom1=atom1,
            atom2=atom2,
            carbon2=carbon2,
            atom3=atom3,
            atom4=atom4,
            bonders=bonders,
            deleters=deleters,
        ),
        atoms=(carbon1, atom1, atom2, carbon2, atom3, atom4),
        bonders=bonders,
        deleters=deleters,
    )



import typing
from dataclasses import dataclass

import stk

T = typing.TypeVar("T", bound=stk.MoleculeRecord)


@dataclass(frozen=True, slots=True)
class CaseData(typing.Generic[T]):
    selector: stk.Selector[T]
    population: dict[T, float]



from functools import partial

import pytest
import stk
from scipy.spatial.distance import euclidean

from ....case_data import CaseData


@pytest.fixture
def monodentate(position, building_block_1):
    point1, point2 = points = (
        position + [-10, 0, 0],
        position + [10, 0, 0],
    )

    def get_fg0_point(building_block):
        return get_closest_point(
            points=points,
            point=get_fg_position(0, building_block),
        )

    def get_core_point(building_block):
        return get_closest_point(
            points=points,
            point=get_core_position(building_block),
        )

    vertex = stk.metal_complex.MonoDentateLigandVertex(
        id=0,
        position=position,
    )

    return CaseData(
        vertex=vertex,
        edges=tuple(get_edge(vertex)),
        building_block=building_block_1,
        position=position,
        alignment_tests={
            get_fg0_point: point1,
            get_core_point: point2,
        },
        functional_group_edges=({0: 0}),
    )


def get_closest_point(points, point):
    return min(points, key=partial(euclidean, point))


def get_fg_position(id, building_block):
    functional_group = next(building_block.get_functional_groups(id))
    return building_block.get_centroid(
        atom_ids=functional_group.get_placer_ids(),
    )


def get_core_position(building_block):
    return building_block.get_centroid(
        atom_ids=building_block.get_core_atom_ids(),
    )


def get_edge(vertex):
    vertex2 = stk.Vertex(1, vertex.get_position() + [-1, 0, 0])
    yield stk.Edge(0, vertex, vertex2)


@pytest.fixture(
    params=(
        lambda: stk.BuildingBlock(
            smiles="c1cc2c(cn1)CCCCC2",
            functional_groups=[
                stk.SmartsFunctionalGroupFactory(
                    smarts="[#6]~[#7X2]~[#6]",
                    bonders=(1,),
                    deleters=(),
                ),
            ],
        ),
    ),
)
def building_block_1(request) -> stk.BuildingBlock:
    return request.param()



import itertools as it

import pytest
import stk

from ..case_data import CaseData


@pytest.fixture
def one_two_reaction(
    functional_group1,
    functional_group2,
    bond_order,
    periodicity,
):
    """
    A :class:`.CaseData` instance.

    """

    return CaseData(
        reaction=stk.OneTwoReaction(
            functional_group1=functional_group1,
            functional_group2=functional_group2,
            bond_order=bond_order,
            periodicity=periodicity,
        ),
        new_atoms=(),
        new_bonds=tuple(
            get_bonds(
                functional_group1=functional_group1,
                functional_group2=functional_group2,
                bond_order=bond_order,
                periodicity=periodicity,
            ),
        ),
        deleted_atoms=tuple(
            it.chain(
                functional_group1.get_deleters(),
                functional_group2.get_deleters(),
            )
        ),
        deleted_bonds=(),
    )


def get_bonds(
    functional_group1,
    functional_group2,
    bond_order,
    periodicity,
):
    bonders1 = functional_group1.get_bonders()
    bonders2 = functional_group2.get_bonders()
    for bonder1, bonder2 in it.product(bonders1, bonders2):
        yield stk.Bond(
            atom1=bonder1,
            atom2=bonder2,
            order=bond_order,
            periodicity=periodicity,
        )



"""
Atom Batch
==========

"""

from stk._internal.atom_info import AtomInfo


class _AtomBatch:
    """
    A batch of atoms.

    """

    __slots__ = ["_atoms", "_atom_infos", "_atom_map", "_positions"]

    def __init__(self, atoms, num_atoms):
        """
        Initialize an :class:`._AtomBatch` instance.

        Parameters
        ----------
        atoms : :class:`iterable` of :class:`.NewAtom`
            The atoms, which should be added to the batch.

        num_atoms : :class:`int`
            The number of atoms in the molecule being constructed,
            before atoms in this batch are taken into account.

        """

        self._atoms = _atoms = []
        self._positions = positions = []
        self._atom_infos = atom_infos = []
        self._atom_map = atom_map = {}

        for id_, (atom, position) in enumerate(atoms, num_atoms):
            _atoms.append(atom.with_id(id_))
            atom_map[atom.get_id()] = _atoms[-1]
            atom_infos.append(
                AtomInfo(
                    atom=_atoms[-1],
                    building_block_atom=None,
                    building_block=None,
                    building_block_id=None,
                )
            )
            positions.append(position)

    def get_positions(self):
        """
        Yield the positions of atoms in the batch.

        Yields
        ------
        :class:`numpy.ndarray`
            The position of an atom in the batch.

        """

        yield from self._positions

    def get_atoms(self):
        """
        Yield the atoms in the batch.

        Yields
        ------
        :class:`.Atom`
            An atom.

        """

        yield from self._atoms

    def get_atom_infos(self):
        """
        Yield info about the atoms in the batch.

        Yields
        ------
        :class:`.AtomInfo`
            Info about an atom in the batch.

        """

        yield from self._atom_infos

    def get_atom_map(self):
        """
        Get a mapping from the old atom id to the new atom.

        Returns
        -------
        :class:`dict`
            Maps the id of an atom provided to the initializer, to
            the new atom held by the batch, which has an updated id.

        """

        return dict(self._atom_map)



import rdkit.Chem.AllChem as rdkit
import stk


def is_equivalent_atom(atom1, atom2):
    assert atom1 is not atom2
    assert atom1.__class__ is atom2.__class__
    assert atom1.get_id() == atom2.get_id()
    assert atom1.get_charge() == atom2.get_charge()
    assert atom1.get_atomic_number() == atom2.get_atomic_number()


_periodic_table = rdkit.GetPeriodicTable()

atomic_numbers = {
    stk.__dict__[
        _periodic_table.GetElementSymbol(atomic_number)
    ]: atomic_number
    for atomic_number in range(1, 119)
}



import typing
from collections.abc import Callable, Iterable
from typing import Any

import numpy as np

from .fitness_normalizer import FitnessNormalizer

T = typing.TypeVar("T")


class Power(FitnessNormalizer[T]):
    """
    Raises fitness values to some power.

    Examples:

        *Raising Fitness Values to a Power*

        Sometimes you might calculate a property for a molecule, where
        that property indicates a low fitness value. You can use
        :class:`.Power` to raise it to the power of -1 to get your
        final fitness value

        .. testcode:: raising-fitness-values-to-a-power

            import stk

            building_block = stk.BuildingBlock('BrCCBr', stk.BromoFactory())
            record1 = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[building_block],
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            )
            record2 = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[building_block],
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            )
            fitness_values = {
                record1: 1,
                record2: 2,
            }
            normalizer = stk.Power(-1)
            normalized_fitness_values = normalizer.normalize(fitness_values)
            assert normalized_fitness_values[record1] == 1
            assert normalized_fitness_values[record2] == 0.5

        *Raising Fitness Values by a Set of Powers*

        In this example, assume that each fitness value consists of a
        :class:`tuple` of numbers, each representing a different property
        of the molecule, and each contributing to the final fitness value.
        The properties can be anything, such as  energy, number of atoms
        or diameter.

        If your final fitness value depends on the combination of these
        properties, you will probably first want to scale them with
        :class:`.DivideByMean`. Once this is done, you may want to
        raise each property by some power. For example
        if you raise the value of one property by ``1`` and another
        by ``-1``, it means that when the value of property 1 is big,
        the fitness value should also be big, but if the value of property
        2 is big, the fitness value should be small.

        Giving a concrete example

        .. testcode:: raising-fitness-values-by-a-set-of-powers

            import stk
            import numpy as np

            building_block = stk.BuildingBlock('BrCCBr', stk.BromoFactory())
            record = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[building_block],
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            )
            fitness_values = {
                record: (2, 2, 2),
            }
            normalizer = stk.Power((1, -1, 2))
            normalized_fitness_values = normalizer.normalize(fitness_values)
            assert np.all(np.equal(
                normalized_fitness_values[record],
                (2, 0.5, 4),
            ))

        *Selectively Normalizing Fitness Values*

        Sometimes, you only want to normalize some members of a population,
        for example if some do not have an assigned fitness value,
        because the fitness calculation failed for whatever reason.
        You can use the `filter` parameter to exclude records from the
        normalization

        .. testcode:: selectively-normalizing-fitness-values

            import stk
            import numpy as np

            building_block = stk.BuildingBlock('BrCCBr', stk.BromoFactory())
            record1 = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[building_block],
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            )
            record2 = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[building_block],
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            )
            fitness_values = {
                record1: (2, 2, 2),
                record2: None,
            }
            normalizer = stk.Power(
                power=(1, -1, 2),
                # Only normalize values which are not None.
                filter=lambda fitness_values, record:
                    fitness_values[record] is not None,
            )
            normalized_fitness_values = normalizer.normalize(fitness_values)
            assert np.all(np.equal(
                normalized_fitness_values[record1],
                (2, 0.5, 4),
            ))
            assert normalized_fitness_values[record2] is None
    """

    def __init__(
        self,
        power: float | Iterable[float],
        filter: Callable[[dict[T, Any], T], bool] = lambda fitness_values,
        record: True,
    ) -> None:
        """
        Parameters:
            power (float | list[float]):
                The power each fitness value is raised to. Can
                be a single number or multiple numbers, depending on the
                form of the fitness value.

            filter:
                A function which returns ``True`` or ``False``. Only
                molecules which return ``True`` will have fitness values
                normalized. By default, all molecules will have fitness
                values normalized.
                The instance passed to the `fitness_values` argument of
                :meth:`.normalize` is passed as the first argument, while
                the second argument will be passed every
                :class:`.MoleculeRecord` in it, one at a time.
        """
        if not isinstance(power, int | float):
            power = tuple(power)
        self._power = power
        self._filter = filter

    def normalize(self, fitness_values: dict[T, Any]) -> dict[T, Any]:
        return {
            record: (
                np.float_power(fitness_value, self._power)
                if self._filter(fitness_values, record)
                else fitness_value
            )
            for record, fitness_value in fitness_values.items()
        }



import pytest
import stk

from ..case_data import CaseData

bb1 = stk.BuildingBlock("BrCCBr", [stk.BromoFactory()])
bb2 = stk.BuildingBlock("BrCC(CBr)CBr", [stk.BromoFactory()])
graph1 = stk.cage.FourPlusSix((bb1, bb2))

bb3 = stk.BuildingBlock("BrCNCBr", [stk.BromoFactory()])
bb4 = stk.BuildingBlock("BrCC(CNCBr)CBr", [stk.BromoFactory()])
graph2 = stk.cage.EightPlusTwelve((bb3, bb4))


@pytest.fixture(
    scope="session",
    params=(
        lambda: CaseData(
            crosser=stk.GeneticRecombination(
                get_gene=stk.BuildingBlock.get_num_functional_groups,
            ),
            records=(
                stk.MoleculeRecord(graph1),
                stk.MoleculeRecord(graph2),
            ),
            crossover_records=(
                stk.CrossoverRecord(
                    molecule_record=stk.MoleculeRecord(
                        topology_graph=graph1,
                    ),
                    crosser_name="GeneticRecombination",
                ),
                stk.CrossoverRecord(
                    molecule_record=stk.MoleculeRecord(
                        topology_graph=graph1.with_building_blocks(
                            building_block_map={bb1: bb3},
                        ),
                    ),
                    crosser_name="GeneticRecombination",
                ),
                stk.CrossoverRecord(
                    molecule_record=stk.MoleculeRecord(
                        topology_graph=graph1.with_building_blocks(
                            building_block_map={bb2: bb4},
                        ),
                    ),
                    crosser_name="GeneticRecombination",
                ),
                stk.CrossoverRecord(
                    molecule_record=stk.MoleculeRecord(
                        topology_graph=graph1.with_building_blocks(
                            building_block_map={
                                bb1: bb3,
                                bb2: bb4,
                            },
                        )
                    ),
                    crosser_name="GeneticRecombination",
                ),
                stk.CrossoverRecord(
                    molecule_record=stk.MoleculeRecord(
                        topology_graph=graph2.with_building_blocks(
                            building_block_map={
                                bb3: bb1,
                                bb4: bb2,
                            },
                        )
                    ),
                    crosser_name="GeneticRecombination",
                ),
                stk.CrossoverRecord(
                    molecule_record=stk.MoleculeRecord(
                        topology_graph=graph2.with_building_blocks(
                            building_block_map={bb4: bb2},
                        )
                    ),
                    crosser_name="GeneticRecombination",
                ),
                stk.CrossoverRecord(
                    molecule_record=stk.MoleculeRecord(
                        topology_graph=graph2.with_building_blocks(
                            building_block_map={bb3: bb1},
                        )
                    ),
                    crosser_name="GeneticRecombination",
                ),
                stk.CrossoverRecord(
                    molecule_record=stk.MoleculeRecord(
                        topology_graph=graph2,
                    ),
                    crosser_name="GeneticRecombination",
                ),
            ),
        ),
    ),
)
def genetic_recombination(request) -> CaseData:
    return request.param()



import itertools as it
from typing import Callable, Iterable

import stk

from ...utilities import is_clone_functional_group


def test_with_ids(
    functional_group: stk.FunctionalGroup,
    get_id_map: Callable[[stk.FunctionalGroup], dict[int, int]],
) -> None:
    """
    Test :meth:`.FunctionalGroup.with_ids`.

    Parameters:

        functional_group:
            The functional group to test.

        get_id_map:
            Takes a single parameter, `functional_group`, and returns a
            valid `id_map` parameter for its
            :meth:`.FunctionalGroup.with_atoms` method. This allows the
            testing of different values of this parameter.

    """

    # Save a clone to ensure that "functional_group" is not changed by
    # the test, because it should be immutable.
    before = functional_group.clone()
    _test_with_atoms(functional_group, get_id_map)
    is_clone_functional_group(before, functional_group)


def _test_with_atoms(
    functional_group: stk.FunctionalGroup,
    get_id_map: Callable[[stk.FunctionalGroup], dict[int, int]],
) -> None:
    """
    Test :meth:`.FunctionalGroup.with_atoms`.

    Parameters:

        functional_group:
            The functional group to test.

        get_id_map:
            Takes a single parameter, `functional_group`, and returns a
            valid `id_map` parameter for its
            :meth:`.FunctionalGroup.with_atoms` method. This allows the
            testing of different values of this parameter.

    """

    id_map = get_id_map(functional_group)
    clone = functional_group.with_ids(id_map)

    is_modified_id_sequence(
        original_ids=functional_group.get_atom_ids(),
        modified_ids=clone.get_atom_ids(),
        id_map=id_map,
    )
    is_modified_id_sequence(
        original_ids=functional_group.get_placer_ids(),
        modified_ids=clone.get_placer_ids(),
        id_map=id_map,
    )
    is_modified_id_sequence(
        original_ids=functional_group.get_core_atom_ids(),
        modified_ids=clone.get_core_atom_ids(),
        id_map=id_map,
    )


def is_modified_id_sequence(
    original_ids: Iterable[int],
    modified_ids: Iterable[int],
    id_map: dict[int, int],
) -> None:
    for original_id, modified_id in it.zip_longest(
        original_ids,
        modified_ids,
    ):
        if original_id in id_map:
            assert modified_id == id_map[original_id]
        else:
            assert original_id == modified_id



import pytest
from pytest_lazyfixture import lazy_fixture

# Fixtures must be visible for lazy_fixture() calls.
from .fixtures import *  # noqa


@pytest.fixture(
    params=(
        lazy_fixture("axle"),
        lazy_fixture("cycle"),
    ),
)
def case_data(request):
    return request.param



import pytest
import stk

from ..case_data import CaseData


def _get_case_data_1() -> CaseData:
    topology_graph = stk.polymer.Linear(
        building_blocks=[
            stk.BuildingBlock("BrCCBr", stk.BromoFactory()),
        ],
        repeating_unit="A",
        num_repeating_units=2,
    )
    return CaseData.new(
        fitness_normalizer=stk.Power(
            power=(0.5, 0, -1),
            filter=lambda fitness_values, record: fitness_values[record]
            is not None,
        ),
        fitness_values={
            stk.MoleculeRecord(
                topology_graph=topology_graph,
            ): ((4, 10, 1), (2, 1, 1)),
            stk.MoleculeRecord(
                topology_graph=topology_graph,
            ): ((9, 20, 2), (3, 1, 0.5)),
            stk.MoleculeRecord(
                topology_graph=topology_graph,
            ): ((16, 30, 4), (4, 1, 0.25)),
            stk.MoleculeRecord(topology_graph): (None, None),
        },
    )


@pytest.fixture(
    scope="session",
    params=(_get_case_data_1,),
)
def power(request: pytest.FixtureRequest) -> CaseData:
    return request.param()



"""
One-Two Reaction
================

"""

import itertools as it

from stk._internal.bond import Bond

from .reaction.reaction import Reaction


class OneTwoReaction(Reaction):
    """
    A reaction between two functional groups.

    One functional group has 1 *bonder* atom, A, and the other has 2, B
    and C. Two bonds are created, one between A and B and one between
    A and C. Any deleter atoms are removed.

    """

    def __init__(
        self,
        functional_group1,
        functional_group2,
        bond_order,
        periodicity,
    ):
        """
        Initialize a :class:`.OneTwoReaction` instance.

        Parameters
        ----------
        functional_group1 : :class:`.GenericFunctionalGroup`
            The first functional group in the reaction.

        functional_group2 : :class:`.GenericFunctionalGroup`
            The second functional group in the reaction.

        bond_order : :class:`int`
            The bond order of the bond created by the reaction.

        periodicity : :class:`tuple` of :class:`int`
            The periodicity of the bond created by the reaction.

        """

        self._functional_group1 = functional_group1
        self._functional_group2 = functional_group2
        self._bond_order = bond_order
        self._periodicity = periodicity

    def _get_new_atoms(self):
        return
        yield

    def _get_new_bonds(self):
        bonders1 = self._functional_group1.get_bonders()
        bonders2 = self._functional_group2.get_bonders()
        for bonder1, bonder2 in it.product(bonders1, bonders2):
            yield Bond(
                atom1=bonder1,
                atom2=bonder2,
                order=self._bond_order,
                periodicity=self._periodicity,
            )

    def _get_deleted_atoms(self):
        yield from self._functional_group1.get_deleters()
        yield from self._functional_group2.get_deleters()

    def _get_deleted_bonds(self):
        return
        yield



from .center import *  # noqa
from .head_1 import *  # noqa
from .head_2 import *  # noqa
from .head_3 import *  # noqa
from .tail_1 import *  # noqa
from .tail_2 import *  # noqa
from .tail_3 import *  # noqa



from stk import AtomInfo


def check_atom_infos(old_state, new_state, building_blocks):
    old_infos = tuple(old_state.get_atom_infos())
    new_infos = tuple(new_state.get_atom_infos())
    num_added = len(new_infos) - len(old_infos)
    assert num_added == sum(bb.get_num_atoms() for bb in building_blocks)

    for info1, info2 in zip(
        # Take just the newly added infos.
        new_infos[len(old_infos) :],
        get_infos(
            building_blocks=building_blocks,
            building_block_index=max(
                (info.building_block_index + 1 for info in old_infos),
                default=0,
            ),
            atom_id=len(old_infos),
        ),
    ):
        is_equivalent_atom(info1.atom, info2.atom)
        assert info1.building_block is info2.building_block
        assert info1.building_block_index is info2.building_block_index


def get_infos(building_blocks, building_block_index, atom_id):
    for building_block in building_blocks:
        for atom in building_block.get_atoms():
            yield AtomInfo(
                atom=atom.with_id(atom_id),
                building_block=building_block,
                building_block_index=building_block_index,
            )
            atom_id += 1
        building_block_index += 1


def is_equivalent_atom(atom1, atom2):
    assert atom1.__class__ is atom2.__class__
    assert atom1.get_id() == atom2.get_id()
    assert atom1.get_charge() == atom2.get_charge()



class CaseData:
    """
    A test case for :class:`.FunctionalGroup` instances.

    Attributes
    ----------
    functional_group : :class:`.FunctionalGroup`
        The functional group to test.

    atoms : :class:`tuple` of :class:`.Atom`
        The correct atoms held by the functional group.

    placers : :class:`tuple` of :class:`.Atom`
        The correct placer atoms held by the functional group.

    core_atoms : :class:`tuple` of :class:`.Atom`
        The correct core atoms held by the functional group.

    """

    def __init__(self, functional_group, atoms, placers, core_atoms):
        """
        Initialize a :class:`.CaseData` instance.

        Parameters
        ----------
        functional_group : :class:`.FunctionalGroup`
            The functional group to test.

        atoms : :class:`tuple` of :class:`.Atom`
            The correct atoms held by the functional group.

        placers : :class:`tuple` of :class:`.Atom`
            The correct placer atoms held by the functional group.

        core_atoms : :class:`tuple` of :class:`.Atom`
            The correct core atoms held by the functional group.

        """

        self.functional_group = functional_group
        self.atoms = atoms
        self.placers = placers
        self.core_atoms = core_atoms


class GenericCaseData(CaseData):
    """
    A test case for :class:`.GenericFunctionalGroup` instances.

    Attributes
    ----------
    functional_group : :class:`.FunctionalGroup`
        The functional group to test.

    atoms : :class:`tuple` of :class:`.Atom`
        The correct atoms held by the functional group.

    placers : :class:`tuple` of :class:`.Atom`
        The correct placer atoms held by the functional group.

    core_atoms : :class:`tuple` of :class:`.Atom`
        The correct core atoms held by the functional group.

    bonders : :class:`tuple` of :class:`.Atom`
        The correct bonder atoms held by the functional group.

    deleters : :class:`tuple` of :class:`.Atom`
        The correct deleter atoms held by the functional group.

    """

    def __init__(self, functional_group, atoms, bonders, deleters):
        """
        Initialize a :class:`.GenericCaseData` instance.

        Parameters
        ----------
        functional_group : :class:`.FunctionalGroup`
            The functional group to test.

        atoms : :class:`tuple` of :class:`.Atom`
            The correct atoms held by `functional_group`.

        bonders : :class:`tuple` of :class:`.Atom`
            The correct bonder atoms of `functional_group`.

        deleters : :class:`tuple` of :class:`.Atom`
            The correct deleter atoms of `functional_group`.

        """

        self.bonders = bonders
        self.deleters = deleters
        deleter_ids = set(atom.get_id() for atom in deleters)
        super().__init__(
            functional_group=functional_group,
            atoms=atoms,
            placers=bonders,
            core_atoms=tuple(
                atom for atom in atoms if atom.get_id() not in deleter_ids
            ),
        )



from .case_data import CaseData

import stk


def test_repr(case_data: CaseData) -> None:
    """
    Test :meth:`.BuildingBlock.__repr__`.

    Parameters:
        case_data:
            A test case. Holds the building block to test.

    """

    _test_repr(
        building_block=case_data.building_block,
        known_repr=case_data.known_repr,
    )


def _test_repr(building_block: stk.BuildingBlock, known_repr: str) -> None:
    """
    Test :meth:`.BuildingBlock.get_functional_groups`.

    Parameters:
        building_block:
            The building block to test.

        known_repr:
            The correct representation.

    """

    print(building_block.__repr__(), known_repr)
    assert building_block.__repr__() == known_repr



def test_get_vertex_ids(case_data):
    """
    Test :meth:`.Edge.get_vertex_ids`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the edge to test and the correct vertex ids.

    Returns
    -------
    None : :class:`NoneType`

    """

    id1, id2 = case_data.edge.get_vertex_ids()
    assert id1 == case_data.vertex1_id
    assert id2 == case_data.vertex2_id



class FunctionalGroupFactory:
    """
    An abstract base class for functional group factories.

    The purpose of a functional group factory is to create
    :class:`.FunctionalGroup` instances. It allows the user to avoid
    creating :class:`.FunctionalGroup` instances manually. The user
    will use :meth:`.get_functional_groups` to achieve this. Subclasses
    of this class are made to customize the automatic creation
    process.

    Examples
    --------
    *Using a Single Functional Group From a Factory*

    You have a building block with two functional groups, but you want
    to use just one.

    .. testcode:: using-a-single-functional-group-factory

        import stk

        amino_factory = stk.PrimaryAminoFactory()
        building_block = stk.BuildingBlock('NCCN')
        amino_group1, *rest = amino_factory.get_functional_groups(
            molecule=building_block,
        )
        building_block = building_block.with_functional_groups(
            functional_groups=(amino_group1, ),
        )

    .. testcode:: using-a-single-functional-group-factory
        :hide:

        assert all(
            isinstance(functional_group, stk.PrimaryAmino)
            for functional_group
            in building_block.get_functional_groups()
        )
        assert building_block.get_num_functional_groups() == 1

    *Using a Subset of Functional Groups From a Factory*

    You have multiple functional groups, but you want the building
    block to use a specific subset.

    .. testcode:: using-a-subset-of-functional-groups-from-a-factory

        import stk

        bromo_factory = stk.BromoFactory()
        building_block = stk.BuildingBlock('BrCC(Br)CC(Br)CC(Br)CCBr')
        bromo_groups = tuple(bromo_factory.get_functional_groups(
            molecule=building_block,
        ))
        building_block = building_block.with_functional_groups(
            functional_groups=(
                bromo_groups[0],
                bromo_groups[3],
                bromo_groups[4],
            ),
        )

    .. testcode:: using-a-subset-of-functional-groups-from-a-factory
        :hide:

        assert all(
            isinstance(functional_group, stk.Bromo)
            for functional_group
            in building_block.get_functional_groups()
        )
        assert building_block.get_num_functional_groups() == 3

    More usage examples can be found in the docstrings of the
    various subclasses.

    *Subclass Implementation*

    The source of the subclasses, listed in
    :mod:`.functional_group_factory`, can serve as good examples.

    """

    def get_functional_groups(self, molecule):
        """
        Yield functional groups in `molecule`.

        Parameters
        ----------
        molecule : :class:`.Molecule`
            The molecule, whose functional groups are to be found.

        Yields
        ------
        :class:`.FunctionalGroup`
            A functional group in `molecule`.

        Examples
        --------
        See :class:`.FunctionalGroupFactory`.

        """

        raise NotImplementedError()

    def __repr__(self) -> str:
        return f"{type(self).__name__}()"



from stk._internal.topology_graphs.host_guest.complex import Complex, Guest
from stk._internal.topology_graphs.host_guest.vertices import (
    GuestVertex,
    HostVertex,
)

__all__ = [
    "Guest",
    "Complex",
    "HostVertex",
    "GuestVertex",
]



import itertools as it

import pytest
import stk

from ..case_data import CaseData
from .utilities import MockConstructionState, MockEdge


@pytest.fixture
def dative_reaction(
    functional_group1,
    functional_group1_2,
    periodicity,
    bond_order,
):
    bond_order_key = frozenset(
        {
            type(functional_group1),
            type(functional_group1_2),
        }
    )
    edge = MockEdge(0, periodicity)
    return CaseData(
        factory=stk.DativeReactionFactory(
            stk.GenericReactionFactory(
                bond_orders={
                    bond_order_key: bond_order,
                },
            )
        ),
        construction_state=MockConstructionState(
            edges=(edge,),
            edge_functional_groups={
                0: (
                    functional_group1,
                    functional_group1_2,
                )
            },
        ),
        edge_group=stk.EdgeGroup(edges=(edge,)),
        reaction_result=stk.ReactionResult(
            new_atoms=(),
            new_bonds=get_new_bonds(
                functional_group1=functional_group1,
                functional_group2=functional_group1_2,
                order=bond_order,
                periodicity=periodicity,
            ),
            deleted_atoms=it.chain(
                functional_group1.get_deleters(),
                functional_group1_2.get_deleters(),
            ),
            deleted_bonds=(),
        ),
    )


def is_metal(atom):
    metal_atomic_numbers = set(
        it.chain(
            range(21, 31),
            range(39, 49),
            range(72, 81),
        )
    )

    return atom.get_atomic_number() in metal_atomic_numbers


def get_new_bonds(
    functional_group1,
    functional_group2,
    order,
    periodicity,
):
    (bonder1,) = functional_group1.get_bonders()
    (bonder2,) = functional_group2.get_bonders()

    if is_metal(bonder1):
        yield stk.Bond(
            atom1=bonder2,
            atom2=bonder1,
            order=order,
            periodicity=periodicity,
        )
    else:
        yield stk.Bond(
            atom1=bonder1,
            atom2=bonder2,
            order=order,
            periodicity=periodicity,
        )



import numpy as np


def is_clone(vertex, clone):
    assert np.all(
        np.equal(
            vertex.get_position(),
            clone.get_position(),
        )
    )
    assert vertex.get_id() == clone.get_id()
    assert np.all(np.equal(vertex.get_cell(), clone.get_cell()))



"""
Primary Amino
=============

"""

from .generic_functional_group import GenericFunctionalGroup


class PrimaryAmino(GenericFunctionalGroup):
    """
    Represents a primary amino functional group.

    The structure of the functional group is given by the pseudo-SMILES
    ``[atom][nitrogen]([hydrogen1])[hydrogen2]``.

    """

    def __init__(
        self,
        nitrogen,
        hydrogen1,
        hydrogen2,
        atom,
        bonders,
        deleters,
        placers=None,
    ):
        """
        Initializes a :class:`.PrimaryAmino` instance.

        Parameters
        ----------
        nitrogen : :class:`.N`
            The ``[nitrogen]`` atom.

        hydrogen1 : :class:`.H`
            The ``[hydrogen1]`` atom.

        hydrogen2 : :class:`.H`
            The ``[hydrogen2]`` atom.

        atom : :class:`.Atom`
            The ``[atom]`` atom.

        bonders : :class:`tuple` of :class:`.Atom`
            The bonder atoms.

        deleters : :class:`tuple` of :class:`.Atom`
            The deleter atoms.

        placers : :class:`tuple` of :class:`.Atom`, optional
            The placer atoms. If ``None`` the `bonders` will be used.

        """

        self._nitrogen = nitrogen
        self._hydrogen1 = hydrogen1
        self._hydrogen2 = hydrogen2
        self._atom = atom
        super().__init__(
            atoms=(nitrogen, hydrogen1, hydrogen2, atom),
            bonders=bonders,
            deleters=deleters,
            placers=bonders if placers is None else placers,
        )

    def get_nitrogen(self):
        """
        Get the ``[nitrogen]`` atom.

        Returns
        -------
        :class:`.N`
            The ``[nitrogen]`` atom.

        """

        return self._nitrogen

    def get_hydrogen1(self):
        """
        Get the ``[hydrogen1]`` atom.

        Returns
        -------
        :class:`.H`
            The ``[hydrogen1]`` atom.

        """

        return self._hydrogen1

    def get_hydrogen2(self):
        """
        Get the ``[hydrogen2]`` atom.

        Returns
        -------
        :class:`.H`
            The ``[hydrogen2]`` atom.

        """

        return self._hydrogen2

    def get_atom(self):
        """
        Get the ``[atom]`` atom.

        Returns
        -------
        :class:`.Atom`
            The ``[atom]`` atom.

        """

        return self._atom

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"{self._nitrogen}, {self._hydrogen1}, {self._hydrogen2}, "
            f"{self._atom}, bonders={self._bonders}, "
            f"deleters={self._deleters}"
            ")"
        )

    def clone(self):
        clone = super().clone()
        clone._nitrogen = self._nitrogen
        clone._hydrogen1 = self._hydrogen1
        clone._hydrogen2 = self._hydrogen2
        clone._atom = self._atom
        return clone

    def with_atoms(self, atom_map):
        clone = super().with_atoms(atom_map)
        clone._nitrogen = atom_map.get(
            self._nitrogen.get_id(),
            self._nitrogen,
        )
        clone._hydrogen1 = atom_map.get(
            self._hydrogen1.get_id(),
            self._hydrogen1,
        )
        clone._hydrogen2 = atom_map.get(
            self._hydrogen2.get_id(),
            self._hydrogen2,
        )
        clone._atom = atom_map.get(
            self._atom.get_id(),
            self._atom,
        )
        return clone



from stk import (
    cage,
    cof,
    host_guest,
    macrocycle,
    metal_complex,
    polymer,
    rotaxane,
    small,
)
from stk._internal.atom import Atom
from stk._internal.atom_info import AtomInfo
from stk._internal.bond import Bond
from stk._internal.bond_info import BondInfo
from stk._internal.building_block import BuildingBlock
from stk._internal.constructed_molecule import ConstructedMolecule
from stk._internal.construction_result.construction_result import (
    ConstructionResult,
)
from stk._internal.construction_state.construction_state import (
    ConstructionState,
)
from stk._internal.construction_state.graph_state import GraphState
from stk._internal.construction_state.molecule_state.molecule_state import (
    MoleculeState,
)
from stk._internal.databases.constructed_molecule import (
    ConstructedMoleculeDatabase,
)
from stk._internal.databases.molecule import MoleculeDatabase
from stk._internal.databases.mongo_db.constructed_molecule import (
    ConstructedMoleculeMongoDb,
)
from stk._internal.databases.mongo_db.molecule import MoleculeMongoDb
from stk._internal.databases.mongo_db.value import ValueMongoDb
from stk._internal.databases.value import ValueDatabase
from stk._internal.ea.crossover.genetic_recombination import (
    GeneticRecombination,
)
from stk._internal.ea.crossover.molecule_crosser import MoleculeCrosser
from stk._internal.ea.crossover.random import RandomCrosser
from stk._internal.ea.crossover.record import CrossoverRecord
from stk._internal.ea.evolutionary_algorithm.evolutionary_algorithm import (
    EvolutionaryAlgorithm,
)
from stk._internal.ea.fitness_calculators.fitness_calculator import (
    FitnessCalculator,
)
from stk._internal.ea.fitness_calculators.fitness_function import (
    FitnessFunction,
)
from stk._internal.ea.fitness_calculators.property_vector import PropertyVector
from stk._internal.ea.fitness_normalizers.add import Add
from stk._internal.ea.fitness_normalizers.divide_by_mean import DivideByMean
from stk._internal.ea.fitness_normalizers.fitness_normalizer import (
    FitnessNormalizer,
)
from stk._internal.ea.fitness_normalizers.multiply import Multiply
from stk._internal.ea.fitness_normalizers.null import NullFitnessNormalizer
from stk._internal.ea.fitness_normalizers.power import Power
from stk._internal.ea.fitness_normalizers.replace_fitness import ReplaceFitness
from stk._internal.ea.fitness_normalizers.sequence import NormalizerSequence
from stk._internal.ea.fitness_normalizers.shift_up import ShiftUp
from stk._internal.ea.fitness_normalizers.sum import Sum
from stk._internal.ea.generation import FitnessValues, Generation
from stk._internal.ea.molecule_record import MoleculeRecord
from stk._internal.ea.mutation.mutator import MoleculeMutator
from stk._internal.ea.mutation.random import RandomMutator
from stk._internal.ea.mutation.random_building_block import RandomBuildingBlock
from stk._internal.ea.mutation.random_topology_graph import RandomTopologyGraph
from stk._internal.ea.mutation.record import MutationRecord
from stk._internal.ea.mutation.similar_building_block import (
    SimilarBuildingBlock,
)
from stk._internal.ea.plotters.progress import ProgressPlotter
from stk._internal.ea.plotters.selection import SelectionPlotter
from stk._internal.ea.selection.batch import Batch, BatchKey
from stk._internal.ea.selection.selectors.above_average import AboveAverage
from stk._internal.ea.selection.selectors.best import Best
from stk._internal.ea.selection.selectors.filter_batches import FilterBatches
from stk._internal.ea.selection.selectors.filter_molecules import (
    FilterMolecules,
)
from stk._internal.ea.selection.selectors.remove_batches import RemoveBatches
from stk._internal.ea.selection.selectors.remove_molecules import (
    RemoveMolecules,
)
from stk._internal.ea.selection.selectors.roulette import Roulette
from stk._internal.ea.selection.selectors.selector import Selector
from stk._internal.ea.selection.selectors.stochastic_universal_sampling import (  # noqa
    StochasticUniversalSampling,
)
from stk._internal.ea.selection.selectors.tournament import Tournament
from stk._internal.ea.selection.selectors.worst import Worst
from stk._internal.elements import (
    Ac,
    Ag,
    Al,
    Am,
    Ar,
    As,
    At,
    Au,
    B,
    Ba,
    Be,
    Bh,
    Bi,
    Bk,
    Br,
    C,
    Ca,
    Cd,
    Ce,
    Cf,
    Cl,
    Cm,
    Cn,
    Co,
    Cr,
    Cs,
    Cu,
    Db,
    Ds,
    Dy,
    Er,
    Es,
    Eu,
    F,
    Fe,
    Fl,
    Fm,
    Fr,
    Ga,
    Gd,
    Ge,
    H,
    He,
    Hf,
    Hg,
    Ho,
    Hs,
    I,
    In,
    Ir,
    K,
    Kr,
    La,
    Li,
    Lr,
    Lu,
    Lv,
    Mc,
    Md,
    Mg,
    Mn,
    Mo,
    Mt,
    N,
    Na,
    Nb,
    Nd,
    Ne,
    Nh,
    Ni,
    No,
    Np,
    O,
    Og,
    Os,
    P,
    Pa,
    Pb,
    Pd,
    Pm,
    Po,
    Pr,
    Pt,
    Pu,
    Ra,
    Rb,
    Re,
    Rf,
    Rg,
    Rh,
    Rn,
    Ru,
    S,
    Sb,
    Sc,
    Se,
    Sg,
    Si,
    Sm,
    Sn,
    Sr,
    Ta,
    Tb,
    Tc,
    Te,
    Th,
    Ti,
    Tl,
    Tm,
    Ts,
    U,
    V,
    W,
    Xe,
    Y,
    Yb,
    Zn,
    Zr,
)
from stk._internal.functional_group_factories.alcohol_factory import (
    AlcoholFactory,
)
from stk._internal.functional_group_factories.aldehyde_factory import (
    AldehydeFactory,
)
from stk._internal.functional_group_factories.amide_factory import (
    AmideFactory,
)
from stk._internal.functional_group_factories.boronic_acid_factory import (
    BoronicAcidFactory,
)
from stk._internal.functional_group_factories.bromo_factory import (
    BromoFactory,
)
from stk._internal.functional_group_factories.carboxylic_acid_factory import (
    CarboxylicAcidFactory,
)
from stk._internal.functional_group_factories.dibromo_factory import (
    DibromoFactory,
)
from stk._internal.functional_group_factories.difluoro_factory import (
    DifluoroFactory,
)
from stk._internal.functional_group_factories.diol_factory import (
    DiolFactory,
)
from stk._internal.functional_group_factories.fluoro_factory import (
    FluoroFactory,
)
from stk._internal.functional_group_factories.functional_group_factory import (
    FunctionalGroupFactory,
)
from stk._internal.functional_group_factories.iodo_factory import (
    IodoFactory,
)
from stk._internal.functional_group_factories.primary_amino_factory import (
    PrimaryAminoFactory,
)
from stk._internal.functional_group_factories.ring_amine_factory import (
    RingAmineFactory,
)
from stk._internal.functional_group_factories.secondary_amino_factory import (
    SecondaryAminoFactory,
)
from stk._internal.functional_group_factories.smarts import (
    SmartsFunctionalGroupFactory,
)
from stk._internal.functional_group_factories.terminal_alkene_factory import (
    TerminalAlkeneFactory,
)
from stk._internal.functional_group_factories.terminal_alkyne_factory import (
    TerminalAlkyneFactory,
)
from stk._internal.functional_group_factories.thioacid_factory import (
    ThioacidFactory,
)
from stk._internal.functional_group_factories.thiol_factory import (
    ThiolFactory,
)
from stk._internal.functional_groups.alcohol import Alcohol
from stk._internal.functional_groups.aldehyde import Aldehyde
from stk._internal.functional_groups.alkene import Alkene
from stk._internal.functional_groups.alkyne import Alkyne
from stk._internal.functional_groups.amide import Amide
from stk._internal.functional_groups.boronic_acid import BoronicAcid
from stk._internal.functional_groups.bromo import Bromo
from stk._internal.functional_groups.carboxylic_acid import CarboxylicAcid
from stk._internal.functional_groups.dibromo import Dibromo
from stk._internal.functional_groups.difluoro import Difluoro
from stk._internal.functional_groups.diol import Diol
from stk._internal.functional_groups.fluoro import Fluoro
from stk._internal.functional_groups.functional_group import FunctionalGroup
from stk._internal.functional_groups.generic_functional_group import (
    GenericFunctionalGroup,
)
from stk._internal.functional_groups.iodo import Iodo
from stk._internal.functional_groups.primary_amino import PrimaryAmino
from stk._internal.functional_groups.ring_amine import RingAmine
from stk._internal.functional_groups.secondary_amino import SecondaryAmino
from stk._internal.functional_groups.single_atom import SingleAtom
from stk._internal.functional_groups.thioacid import Thioacid
from stk._internal.functional_groups.thiol import Thiol
from stk._internal.json_serde.constructed_molecule import (
    ConstructedMoleculeDejsonizer,
    ConstructedMoleculeJsonizer,
)
from stk._internal.json_serde.molecule import (
    MoleculeDejsonizer,
    MoleculeJsonizer,
)
from stk._internal.key_makers.inchi import Inchi
from stk._internal.key_makers.inchi_key import InchiKey
from stk._internal.key_makers.molecule import MoleculeKeyMaker
from stk._internal.key_makers.smiles import Smiles
from stk._internal.molecule import Molecule
from stk._internal.optimizers.collapser import Collapser
from stk._internal.optimizers.mchammer import MCHammer
from stk._internal.optimizers.null import NullOptimizer
from stk._internal.optimizers.optimizer import Optimizer
from stk._internal.optimizers.periodic_collapser import PeriodicCollapser
from stk._internal.optimizers.spinner import Spinner
from stk._internal.periodic_info import PeriodicInfo
from stk._internal.reaction_factories.dative_reaction_factory import (
    DativeReactionFactory,
)
from stk._internal.reaction_factories.generic_reaction_factory import (
    GenericReactionFactory,
)
from stk._internal.reaction_factories.reaction_factory import ReactionFactory
from stk._internal.reactions.dative_reaction.dative_reaction import (
    DativeReaction,
)
from stk._internal.reactions.one_one_reaction import OneOneReaction
from stk._internal.reactions.one_two_reaction import OneTwoReaction
from stk._internal.reactions.reaction.reaction import Reaction
from stk._internal.reactions.reaction.reaction_result import ReactionResult
from stk._internal.reactions.ring_amine_reaction import RingAmineReaction
from stk._internal.reactions.two_two_reaction import TwoTwoReaction
from stk._internal.topology_graphs.edge import Edge
from stk._internal.topology_graphs.edge_group import EdgeGroup
from stk._internal.topology_graphs.topology_graph.topology_graph import (
    TopologyGraph,
)
from stk._internal.topology_graphs.vertex import Vertex
from stk._internal.utilities.utilities import (
    get_acute_vector,
    normalize_vector,
    rotation_matrix_arbitrary_axis,
    vector_angle,
)
from stk._internal.writers.mdl_mol import MolWriter
from stk._internal.writers.pdb import PdbWriter
from stk._internal.writers.turbomole import TurbomoleWriter
from stk._internal.writers.xyz import XyzWriter
from stk._version import __version__

BatchKey = BatchKey
"""A unique key for a :class:`.Batch`."""

__all__ = [
    "Atom",
    "AtomInfo",
    "Bond",
    "BondInfo",
    "MoleculeRecord",
    "EvolutionaryAlgorithm",
    "MoleculeMutator",
    "RandomMutator",
    "MoleculeState",
    "Edge",
    "Vertex",
    "cof",
    "host_guest",
    "metal_complex",
    "get_acute_vector",
    "normalize_vector",
    "rotation_matrix_arbitrary_axis",
    "vector_angle",
    "CrossoverRecord",
    "ReplaceFitness",
    "RemoveMolecules",
    "MutationRecord",
    "NormalizerSequence",
    "SimilarBuildingBlock",
    "RandomBuildingBlock",
    "RandomTopologyGraph",
    "GeneticRecombination",
    "Generation",
    "PropertyVector",
    "AboveAverage",
    "Best",
    "Worst",
    "FilterBatches",
    "StochasticUniversalSampling",
    "Tournament",
    "RemoveBatches",
    "FilterMolecules",
    "Roulette",
    "Batch",
    "BatchKey",
    "Add",
    "NullFitnessNormalizer",
    "NullOptimizer",
    "Power",
    "DivideByMean",
    "Sum",
    "Multiply",
    "ShiftUp",
    "MoleculeJsonizer",
    "MoleculeDejsonizer",
    "ConstructedMoleculeDejsonizer",
    "ConstructedMoleculeJsonizer",
    "Spinner",
    "MoleculeDejsonizer",
    "MoleculeJsonizer",
    "BuildingBlock",
    "ConstructedMolecule",
    "TopologyGraph",
    "MoleculeCrosser",
    "Optimizer",
    "RandomCrosser",
    "FitnessCalculator",
    "FitnessFunction",
    "FitnessNormalizer",
    "AlcoholFactory",
    "AldehydeFactory",
    "AmideFactory",
    "BoronicAcidFactory",
    "BromoFactory",
    "CarboxylicAcidFactory",
    "DibromoFactory",
    "DifluoroFactory",
    "DiolFactory",
    "FunctionalGroupFactory",
    "FluoroFactory",
    "IodoFactory",
    "PrimaryAminoFactory",
    "RingAmineFactory",
    "SecondaryAminoFactory",
    "SmartsFunctionalGroupFactory",
    "TerminalAlkeneFactory",
    "TerminalAlkyneFactory",
    "ThioacidFactory",
    "ThiolFactory",
    "Alcohol",
    "Aldehyde",
    "Alkene",
    "Alkyne",
    "Amide",
    "BoronicAcid",
    "Bromo",
    "CarboxylicAcid",
    "Dibromo",
    "Difluoro",
    "Diol",
    "Fluoro",
    "FunctionalGroup",
    "GenericFunctionalGroup",
    "Iodo",
    "PrimaryAmino",
    "RingAmine",
    "RingAmineReaction",
    "SecondaryAmino",
    "SingleAtom",
    "Thioacid",
    "Thiol",
    "cage",
    "polymer",
    "small",
    "Collapser",
    "MCHammer",
    "ReactionFactory",
    "Reaction",
    "Selector",
    "ProgressPlotter",
    "SelectionPlotter",
    "XyzWriter",
    "PdbWriter",
    "MolWriter",
    "TurbomoleWriter",
    "DativeReactionFactory",
    "GenericReactionFactory",
    "ValueDatabase",
    "ConstructedMoleculeMongoDb",
    "ConstructedMoleculeDatabase",
    "MoleculeMongoDb",
    "ValueMongoDb",
    "MoleculeDatabase",
    "Inchi",
    "InchiKey",
    "Smiles",
    "MoleculeKeyMaker",
    "Molecule",
    "ConstructionState",
    "ConstructionResult",
    "ReactionResult",
    "GraphState",
    "OneOneReaction",
    "OneTwoReaction",
    "TwoTwoReaction",
    "DativeReaction",
    "PeriodicInfo",
    "rotaxane",
    "macrocycle",
    "EdgeGroup",
    "PeriodicCollapser",
    "FitnessValues",
    "__version__",
    "Ac",
    "Ag",
    "Al",
    "Am",
    "Ar",
    "As",
    "At",
    "Au",
    "B",
    "Ba",
    "Be",
    "Bh",
    "Bi",
    "Bk",
    "Br",
    "C",
    "Ca",
    "Cd",
    "Ce",
    "Cf",
    "Cl",
    "Cm",
    "Cn",
    "Co",
    "Cr",
    "Cs",
    "Cu",
    "Db",
    "Ds",
    "Dy",
    "Er",
    "Es",
    "Eu",
    "F",
    "Fe",
    "Fl",
    "Fm",
    "Fr",
    "Ga",
    "Gd",
    "Ge",
    "H",
    "He",
    "Hf",
    "Hg",
    "Ho",
    "Hs",
    "I",
    "In",
    "Ir",
    "K",
    "Kr",
    "La",
    "Li",
    "Lr",
    "Lu",
    "Lv",
    "Mc",
    "Md",
    "Mg",
    "Mn",
    "Mo",
    "Mt",
    "N",
    "Na",
    "Nb",
    "Nd",
    "Ne",
    "Nh",
    "Ni",
    "No",
    "Np",
    "O",
    "Og",
    "Os",
    "P",
    "Pa",
    "Pb",
    "Pd",
    "Pm",
    "Po",
    "Pr",
    "Pt",
    "Pu",
    "Ra",
    "Rb",
    "Re",
    "Rf",
    "Rg",
    "Rh",
    "Rn",
    "Ru",
    "S",
    "Sb",
    "Sc",
    "Se",
    "Sg",
    "Si",
    "Sm",
    "Sn",
    "Sr",
    "Ta",
    "Tb",
    "Tc",
    "Te",
    "Th",
    "Ti",
    "Tl",
    "Tm",
    "Ts",
    "U",
    "V",
    "W",
    "Xe",
    "Y",
    "Yb",
    "Zn",
    "Zr",
]



import typing
from collections.abc import Iterator

from stk._internal.ea.molecule_record import MoleculeRecord
from stk._internal.ea.selection.batch import Batch

from .selector import ExcludedBatches, IncludedBatches, Selector

T = typing.TypeVar("T", bound=MoleculeRecord)


class RemoveMolecules(Selector[T]):
    """
    Prevents a :class:`.Selector` from selecting some molecules.

    Examples:

        *Removing Molecules From Selection*

        .. testcode:: removing-molecules-from-selection

            import stk

            selector = stk.RemoveMolecules(
                # Do not select any molecules from the top 5 batches
                # of size 3.
                remover=stk.Best(num_batches=5, batch_size=3),
                # Select the ret of the molecules using roulette.
                selector=stk.Roulette(num_batches=20, batch_size=3),
            )
            population = {
                stk.MoleculeRecord(
                    topology_graph=stk.polymer.Linear(
                        building_blocks=[
                            stk.BuildingBlock('BrCCBr', stk.BromoFactory()),
                        ],
                        repeating_unit='A',
                        num_repeating_units=2,
                    ),
                ): i
                for i in range(10)
            }
            for batch in selector.select(population):
                # Do stuff with batch. The batch is guaranteed not to
                # contain any molecules which are found in the best 5
                # batches of size 3.
                pass
    """

    def __init__(self, remover: Selector[T], selector: Selector[T]) -> None:
        """
        Parameters:
            remover:
                Selects batches molecules, any molecule selected cannot
                be selected by `selector`.

            selector:
                Selects batches of molecules, not containing any molecules
                selected by `remover`.
        """
        self._remover = remover
        self._selector = selector

    def select(
        self,
        population: dict[T, float],
        included_batches: "IncludedBatches" = None,
        excluded_batches: "ExcludedBatches" = None,
    ) -> Iterator[Batch[T]]:
        remover_batches = self._remover.select(
            population=population,
            included_batches=included_batches,
            excluded_batches=excluded_batches,
        )
        removed = {record for batch in remover_batches for record in batch}
        valid_population = {
            record: population[record]
            for record in population
            if record not in removed
        }
        yield from self._selector.select(
            population=valid_population,
            included_batches=included_batches,
            excluded_batches=excluded_batches,
        )



import itertools as it

from .molecule import (
    is_equivalent_atom,
    is_equivalent_bond,
    is_equivalent_molecule,
)


def is_equivalent_constructed_molecule(
    constructed_molecule1,
    constructed_molecule2,
):
    is_equivalent_molecule(
        molecule1=constructed_molecule1,
        molecule2=constructed_molecule2,
    )
    are_equivalent_atom_infos(
        infos1=constructed_molecule1.get_atom_infos(),
        infos2=constructed_molecule2.get_atom_infos(),
    )
    are_equivalent_bond_infos(
        infos1=constructed_molecule1.get_bond_infos(),
        infos2=constructed_molecule2.get_bond_infos(),
    )

    for building_block1, building_block2 in it.zip_longest(
        constructed_molecule1.get_building_blocks(),
        constructed_molecule2.get_building_blocks(),
    ):
        is_equivalent_molecule(
            building_block1,
            building_block2,
        )

    for building_block1, building_block2 in zip(
        constructed_molecule1.get_building_blocks(),
        constructed_molecule2.get_building_blocks(),
    ):
        num1 = constructed_molecule1.get_num_building_block(
            building_block=building_block1,
        )
        num2 = constructed_molecule2.get_num_building_block(
            building_block=building_block2,
        )
        assert num1 == num2


def are_equivalent_atom_infos(infos1, infos2):
    for info1, info2 in it.zip_longest(infos1, infos2):
        is_equivalent_atom_info(info1, info2)


def are_equivalent_bond_infos(infos1, infos2):
    for info1, info2 in it.zip_longest(infos1, infos2):
        is_equivalent_bond_info(info1, info2)


def is_equivalent_atom_info(info1, info2):
    is_equivalent_atom(info1.get_atom(), info2.get_atom())
    if info1.get_building_block() is None:
        assert info2.get_building_block() is None
    else:
        is_equivalent_molecule(
            info1.get_building_block(),
            info2.get_building_block(),
        )

    if info1.get_building_block_atom() is None:
        assert info2.get_building_block_atom() is None
    else:
        is_equivalent_atom(
            atom1=info1.get_building_block_atom(),
            atom2=info2.get_building_block_atom(),
        )

    assert info1.get_building_block_id() == info2.get_building_block_id()


def is_equivalent_bond_info(info1, info2):
    is_equivalent_bond(info1.get_bond(), info2.get_bond())
    if info1.get_building_block() is None:
        assert info2.get_building_block() is None
    else:
        is_equivalent_molecule(
            info1.get_building_block(),
            info2.get_building_block(),
        )
    assert info1.get_building_block_id() == info2.get_building_block_id()



"""
Dative Reaction Factory
=======================

"""

from ..reactions.dative_reaction.dative_reaction import DativeReaction
from .reaction_factory import ReactionFactory


class DativeReactionFactory(ReactionFactory):
    """
    Create :class:`.DativeReaction` instances.

    This reaction factory assumes that the functional groups, which
    belong to the :class:`.EdgeGroup` passed to it, are
    :class:`.GenericFunctionalGroup` instances. It returns a
    :class:`.Reaction` suitable for two such instances.

    Dative bonds are defined with a `bond_order` of 9,
    running from the non-metal atom to the metal atom.

    """

    def __init__(self, reaction_factory):
        """
        Initialize a :class:`.DativeReactionFactory`.

        Parameters
        ----------
        reaction_factory : :class:`.GenericReactionFactory`
            Used to create reactions.

        """

        self._reaction_factory = reaction_factory

    def get_reaction(self, construction_state, edge_group):
        reaction = self._reaction_factory.get_reaction(
            construction_state=construction_state,
            edge_group=edge_group,
        )
        return DativeReaction(reaction)

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return f"{self.__class__.__name__}({self._reaction_factory})"



import itertools
import typing
from collections.abc import Callable, Iterable, Iterator, Sequence

import numpy as np

from stk._internal.ea.molecule_record import MoleculeRecord
from stk._internal.ea.selection.batch import Batch
from stk._internal.ea.selection.selectors.yielded_batches import YieldedBatches
from stk._internal.key_makers.inchi import Inchi
from stk._internal.key_makers.molecule import MoleculeKeyMaker

from .selector import Selector

T = typing.TypeVar("T", bound=MoleculeRecord)


class AboveAverage(Selector[T]):
    """
    Yields above average batches of molecules.

    Examples:

        *Yielding Single Molecule Batches*

        Yielding molecules one at a time. For example, if molecules need
        to be selected for mutation or the next generation

        .. testcode:: yielding-single-molecule-batches

            import stk

            # Make the selector.
            above_avg = stk.AboveAverage()

            population = {
                stk.MoleculeRecord(
                    topology_graph=stk.polymer.Linear(
                        building_blocks=[
                            stk.BuildingBlock('BrCCBr', stk.BromoFactory()),
                        ],
                        repeating_unit='A',
                        num_repeating_units=2,
                    ),
                ): 1,
                stk.MoleculeRecord(
                    topology_graph=stk.polymer.Linear(
                        building_blocks=[
                            stk.BuildingBlock('BrCCBr', stk.BromoFactory()),
                        ],
                        repeating_unit='A',
                        num_repeating_units=2,
                    ),
                ): 2,
            }

            # Select the molecules.
            for selected, in above_avg.select(population):
                # Do stuff with each selected molecule.
                pass

        *Yielding Batches Holding Multiple Molecules*

        Yielding multiple molecules at once. For example, if molecules need
        to be selected for crossover.

        .. testcode:: yielding-batches-holding-multiple-molecules

            import stk

            # Make the selector.
            above_avg = stk.AboveAverage(batch_size=2)

            population = {
                stk.MoleculeRecord(
                    topology_graph=stk.polymer.Linear(
                        building_blocks=[
                            stk.BuildingBlock('BrCCBr', stk.BromoFactory()),
                        ],
                        repeating_unit='A',
                        num_repeating_units=2,
                    ),
                ): 1,
                stk.MoleculeRecord(
                    topology_graph=stk.polymer.Linear(
                        building_blocks=[
                            stk.BuildingBlock('BrCCBr', stk.BromoFactory()),
                        ],
                        repeating_unit='A',
                        num_repeating_units=2,
                    ),
                ): 2,
            }

            # Select the molecules.
            for selected1, selected2 in above_avg.select(population):
                # Do stuff with the selected molecules.
                pass
    """

    def __init__(
        self,
        num_batches: int | None = None,
        batch_size: int = 1,
        duplicate_molecules: bool = True,
        duplicate_batches: bool = True,
        key_maker: MoleculeKeyMaker = Inchi(),
        fitness_modifier: Callable[
            [dict[T, float]], dict[T, float]
        ] = lambda x: x,
    ) -> None:
        """
        Parameters:
            num_batches:
                The number of batches to yield. If ``None`` then yielding
                will continue forever or until the generator is exhausted,
                whichever comes first.

            batch_size:
                The number of molecule records in each yielded
                :class:`.Batch`.

            duplicate_molecules:
                If ``True`` the same molecule can be yielded in more
                than one batch.

            duplicate_batches:
                If ``True`` the same batch can be yielded more than once.

            key_maker:
                Used to get the keys of molecules. If two molecules have
                the same key, they are considered duplicates.

            fitness_modifier:
                Takes the `population` on which :meth:`~Selector.select`
                is called and returns a :class:`dict`, which maps records
                in the `population` to the fitness values the
                :class:`.Selector` should use.
        """
        super().__init__(key_maker, fitness_modifier, batch_size)
        self._duplicate_molecules = duplicate_molecules
        self._duplicate_batches = duplicate_batches
        self._num_batches = num_batches

    def _select_from_batches(
        self,
        batches: Sequence[Batch[T]],
        yielded_batches: YieldedBatches[T],
    ) -> Iterator[Batch[T]]:
        mean = np.mean([batch.get_fitness_value() for batch in batches])
        # Yield highest fitness batches first.
        selected_batches: Iterable[Batch[T]] = sorted(batches, reverse=True)
        # Yield only batches with a fitness larger than the mean.
        selected_batches = itertools.takewhile(
            lambda batch: batch.get_fitness_value() > mean, selected_batches
        )
        # Yield batches which are multiple times better than the mean
        # multiple times.
        selected_batches = (
            batch
            for batch in selected_batches
            for _ in range(self._get_num_duplicates(batch, float(mean)))
        )
        # If duplicate molecules are not allowed, allow only
        # batches with no yielded molecules.
        if not self._duplicate_molecules:
            selected_batches = filter(
                yielded_batches.has_no_yielded_molecules,
                selected_batches,
            )
        # If duplicate batches are not allowed, allow only
        # unyielded batches.
        if not self._duplicate_batches:
            selected_batches = filter(
                yielded_batches.is_unyielded_batch,
                selected_batches,
            )
        # Limit the number of yielded batches to _num_batches.
        yield from itertools.islice(selected_batches, self._num_batches)

    def _get_num_duplicates(self, batch: Batch[T], mean: float) -> int:
        if self._duplicate_batches and self._duplicate_molecules:
            return int(batch.get_fitness_value() // mean)
        return 1



import pytest
import stk
from pytest_lazyfixture import lazy_fixture

from .case_data import CaseData
from .utilities import atomic_numbers


@pytest.fixture(
    params=[
        cls
        for cls in stk.__dict__.values()
        if isinstance(cls, type)
        and issubclass(cls, stk.Atom)
        and cls is not stk.Atom
    ],
)
def cls(request):
    """
    Return an :class:`.Atom` instance.

    Parameters
    ----------
    id : :class:`int`
        The id of the returned atom.

    charge : :class:`float`
        The charge of the returned atom.

    Returns
    -------
    :class:`.Atom`
        An atom.

    """

    return request.param


@pytest.fixture
def case_data_1(atomic_number, id, charge):
    """
    A :class:`.CaseData` instance.

    """

    return CaseData(
        atom=stk.Atom(
            id=id,
            atomic_number=atomic_number,
            charge=charge,
        ),
        id=id,
        charge=charge,
        atomic_number=atomic_number,
    )


@pytest.fixture
def case_data_2(cls, id, charge):
    """
    A :class:`.CaseData` instance.

    """

    return CaseData(
        atom=cls(id, charge),
        id=id,
        charge=charge,
        atomic_number=atomic_numbers[cls],
    )


@pytest.fixture(
    params=(
        lazy_fixture("case_data_1"),
        lazy_fixture("case_data_2"),
    )
)
def case_data(request):
    """
    A :class:`.CaseData` instance.

    """

    return request.param


@pytest.fixture
def atom(cls):
    """
    An :class:`.Atom` instance.

    """

    return cls(3, -5).clone()


@pytest.fixture(
    params=[0, 3],
)
def id(request):
    """
    An atom id.

    """

    return request.param


@pytest.fixture(
    params=[0],
)
def charge(request):
    """
    An atomic charge.

    """

    return request.param


@pytest.fixture(params=tuple(range(1, 118)))
def atomic_number(request):
    """
    An atomic number.

    """

    return request.param



import pytest
import stk

from ...case_data import CaseData


@pytest.fixture(
    scope="session",
    params=(
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.rotaxane.NRotaxane(
                    axle=_get_axle(),
                    cycles=(
                        _get_cycle1(),
                        _get_cycle2(),
                    ),
                    repeating_unit="AB",
                    num_repeating_units=3,
                ),
            ),
            smiles=(
                "BrC1=C(C2=C(C3=C(C4=C(C5=C(C6=C(C7=C(C8=C(C9=C(C%10=C"
                "(C%11=C(C%12=C(C%13=C(C%14=C(C%15=C(Br)N=[C+]%15)N=[C"
                "+]%14)N=[C+]%13)N=[C+]%12)N=[C+]%11)N=[C+]%10)N=[C+]9"
                ")N=[C+]8)N=[C+]7)N=[C+]6)N=[C+]5)N=[C+]4)N=[C+]3)N=[C"
                "+]2)N=[C+]1.N1=NN2N1N1N=NN1N1N=NN1N1N=NN1N1N=NN1N1N=N"
                "N1N1N=NN1N1N=NN21.N1=NN2N1N1N=NN1N1N=NN1N1N=NN1N1N=NN"
                "1N1N=NN1N1N=NN1N1N=NN21.N1=NN2N1N1N=NN1N1N=NN1N1N=NN1"
                "N1N=NN1N1N=NN1N1N=NN1N1N=NN21.[C+]1=[C+][C+]2[C+]1[C+"
                "]1[C+]=[C+][C+]1[C+]1[C+]=[C+][C+]1[C+]1[C+]=[C+][C+]"
                "1[C+]1[C+]=[C+][C+]1[C+]1[C+]=[C+][C+]1[C+]1[C+]=[C+]"
                "[C+]1[C+]1[C+]=[C+][C+]21.[C+]1=[C+][C+]2[C+]1[C+]1[C"
                "+]=[C+][C+]1[C+]1[C+]=[C+][C+]1[C+]1[C+]=[C+][C+]1[C+"
                "]1[C+]=[C+][C+]1[C+]1[C+]=[C+][C+]1[C+]1[C+]=[C+][C+]"
                "1[C+]1[C+]=[C+][C+]21.[C+]1=[C+][C+]2[C+]1[C+]1[C+]=["
                "C+][C+]1[C+]1[C+]=[C+][C+]1[C+]1[C+]=[C+][C+]1[C+]1[C"
                "+]=[C+][C+]1[C+]1[C+]=[C+][C+]1[C+]1[C+]=[C+][C+]1[C+"
                "]1[C+]=[C+][C+]21"
            ),
            name=name,
        ),
    ),
)
def rotaxane_nrotaxane(request) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )


def _get_axle() -> stk.BuildingBlock:
    molecule = stk.ConstructedMolecule(
        topology_graph=stk.polymer.Linear(
            building_blocks=(
                stk.BuildingBlock(
                    smiles="BrC1=C(Br)[C+]=N1",
                    functional_groups=[stk.BromoFactory()],
                ),
                stk.BuildingBlock(
                    smiles="Br[C+]=NC#CBr",
                    functional_groups=[stk.BromoFactory()],
                ),
            ),
            repeating_unit="A",
            num_repeating_units=15,
        ),
    )
    return stk.BuildingBlock.init_from_molecule(molecule)


def _get_cycle1() -> stk.BuildingBlock:
    molecule = stk.ConstructedMolecule(
        topology_graph=stk.macrocycle.Macrocycle(
            building_blocks=(
                stk.BuildingBlock(
                    smiles="BrN1N(Br)N=N1",
                    functional_groups=[stk.BromoFactory()],
                ),
            ),
            repeating_unit="A",
            num_repeating_units=8,
        ),
    )

    return stk.BuildingBlock.init_from_molecule(molecule)


def _get_cycle2() -> stk.BuildingBlock:
    molecule = stk.ConstructedMolecule(
        topology_graph=stk.macrocycle.Macrocycle(
            building_blocks=(
                stk.BuildingBlock(
                    smiles="Br[C+]1[C+](Br)[C+]=[C+]1",
                    functional_groups=[stk.BromoFactory()],
                ),
            ),
            repeating_unit="A",
            num_repeating_units=8,
        ),
    )
    return stk.BuildingBlock.init_from_molecule(molecule)



from ..utilities import is_clone_generic_functional_group


def test_clone(generic_functional_group):
    """
    Test :meth:`.GenericFunctionalGroup.clone`.

    Parameters
    ----------
    generic_functional_group : :class:`.GenericFunctionalGroup`
        The functional group to test.

    Returns
    -------
    None : :class:`NoneType`

    """

    clone = generic_functional_group.clone()
    is_clone_generic_functional_group(generic_functional_group, clone)



import typing
from collections.abc import Iterator

from stk._internal.ea.molecule_record import MoleculeRecord
from stk._internal.ea.selection.batch import Batch

from .selector import ExcludedBatches, IncludedBatches, Selector

T = typing.TypeVar("T", bound=MoleculeRecord)


class RemoveBatches(Selector[T]):
    """
    Prevents a :class:`.Selector` from selecting some batches.

    Examples:
        *Removing Batches From Selection*

        You want to use :class:`.Roulette` selection on all but the
        5 :class:`.Worst` batches

        .. testcode:: removing-batches-from-selection

            import stk

            selector = stk.RemoveBatches(
                remover=stk.Worst(5),
                selector=stk.Roulette(20),
            )
            population = {
                stk.MoleculeRecord(
                    topology_graph=stk.polymer.Linear(
                        building_blocks=[
                            stk.BuildingBlock('BrCCBr', stk.BromoFactory()),
                        ],
                        repeating_unit='A',
                        num_repeating_units=2,
                    ),
                ): i
                for i in range(100)
            }
            for batch in selector.select(population):
                # Do stuff with batch. It was selected with roulette
                # selection and is not one of the worst 5 batches.
                pass
    """

    def __init__(self, remover: Selector[T], selector: Selector[T]) -> None:
        """
        Parameters:
            remover : :class:`.Selector`
                Selects batches of molecules, which cannot be yielded by
                `selector`.

            selector : :class:`.Selector`
                Selects batches of molecules, except those selected by
                `remover`.
        """
        self._remover = remover
        self._selector = selector

    def select(
        self,
        population: dict[T, float],
        included_batches: "IncludedBatches" = None,
        excluded_batches: "ExcludedBatches" = None,
    ) -> Iterator[Batch[T]]:
        removed_batches = {
            batch.get_identity_key()
            for batch in self._remover.select(
                population=population,
                included_batches=included_batches,
                excluded_batches=excluded_batches,
            )
        }
        if excluded_batches is not None:
            removed_batches |= excluded_batches

        yield from self._selector.select(
            population=population,
            included_batches=included_batches,
            excluded_batches=removed_batches,
        )



import numpy as np

from ..utilities import get_centroid


def get_plane_normal(position_matrix, atom_ids):
    # In this case, Molecule.get_plane_normal should raise the error,
    # so this should should just fail silently, and the explicit
    # failure of Molecule.get_plane_normal will be caught.
    if len(atom_ids) == 0:
        return

    atomic_positions = position_matrix[atom_ids, :]
    centroid = get_centroid(position_matrix, atom_ids)
    return np.around(
        a=np.linalg.svd(atomic_positions - centroid)[-1][2, :],
        decimals=14,
    )



from dataclasses import dataclass

import stk


@dataclass(slots=True, frozen=True)
class CaseData:
    molecules: list[stk.Molecule]
    property_dicts: list[dict]
    expected_count: int
    name: str



"""
MDL Writing Utilities
=====================

"""


def _write_mdl_mol_file(self, path, atom_ids):
    """
    Write to a V3000 ``.mol`` file.

    This function should not be used directly, only via
    :meth:`write`.

    Parameters
    ----------
    path : :class:`str`
        The full path to the file being written.

    atom_ids : :class:`iterable` of :class:`int`
        The atom ids of atoms to write. Can be a single
        :class:`int`, if a single atom is to be used, or ``None``,
        if all atoms are to be used.

    Returns
    -------
    None : :class:`NoneType`

    """

    with open(path, "w") as f:
        f.write(_to_mdl_mol_block(self, atom_ids))


def _to_mdl_mol_block(self, atom_ids=None):
    """
    Return a V3000 mol block of the molecule.

    Parameters
    ----------
    atom_ids : :class:`iterable` of :class:`int`, optional
        The atom ids of atoms to write. Can be a single
        :class:`int`, if a single atom is to be used, or
        ``None``, if all atoms are to be used.

    Returns
    -------
    :class:`str`
        The V3000 mol block representing the molecule.

    """

    if atom_ids is None:
        atom_ids = range(len(self._atoms))
    elif isinstance(atom_ids, int):
        atom_ids = (atom_ids,)

    atom_lines = []
    # This set gets used by bonds.
    id_map = {}
    for new_atom_id, old_atom_id in enumerate(atom_ids, 1):
        id_map[old_atom_id] = new_atom_id

        x, y, z = self._position_matrix[:, old_atom_id]
        atom = self._atoms[old_atom_id]
        symbol = atom.__class__.__name__
        charge = atom.get_charge()
        charge = f" CHG={charge}" if charge else ""
        atom_lines.append(
            f"M  V30 {new_atom_id} {symbol} {x:.4f} "
            f"{y:.4f} {z:.4f} 0{charge}\n"
        )
    atom_block = "".join(atom_lines)

    bond_lines = []
    for bond in self._bonds:
        a1 = bond.get_atom1().get_id()
        a2 = bond.get_atom2().get_id()
        if a1 in id_map and a2 in id_map:
            bond_lines.append(
                f"M  V30 {len(bond_lines) + 1} "
                f"{int(bond.get_order())} "
                f"{id_map[a1]} {id_map[a2]}\n"
            )

    num_bonds = len(bond_lines)
    bond_block = "".join(bond_lines)
    return (
        "\n"
        "     RDKit          3D\n"
        "\n"
        "  0  0  0  0  0  0  0  0  0  0999 V3000\n"
        "M  V30 BEGIN CTAB\n"
        f"M  V30 COUNTS {len(id_map)} {num_bonds} 0 0 0\n"
        "M  V30 BEGIN ATOM\n"
        f"{atom_block}"
        "M  V30 END ATOM\n"
        "M  V30 BEGIN BOND\n"
        f"{bond_block}"
        "M  V30 END BOND\n"
        "M  V30 END CTAB\n"
        "M  END\n"
        "\n"
        "$$$$\n"
    )



def test_get_periodicity(case_data):
    """
    Test :meth:`.Edge.get_periodicity`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the edge to test and the correct
        periodicity.

    Returns
    -------
    None : :class:`NoneType`

    """

    assert case_data.edge.get_periodicity() == case_data.periodicity



import pytest
import stk

from ....case_data import CaseData


@pytest.fixture(
    scope="session",
    params=(
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.cage.SixPlusNine(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles="BrC1=C(Br)[C+]=N1",
                            functional_groups=[stk.BromoFactory()],
                        ),
                        stk.BuildingBlock(
                            smiles=(
                                "Br[C+]1[C+2][C+](Br)[C+]2[C+](F)[C+2]"
                                "C2(Br)[C+2]1"
                            ),
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                ),
            ),
            smiles=(
                "F[C+]1[C+2]C23[C+2][C+]4[C+2][C+](C5=C(N=[C+]5)[C+]5["
                "C+2][C+]6[C+2]C7([C+2][C+](F)[C+]57)C5=C(N=[C+]5)[C+]"
                "5[C+2][C+]([C+2]C7([C+2][C+](F)[C+]57)C5=C4[C+]=N5)C4"
                "=C(N=[C+]4)[C+]4[C+2][C+]5C7=C([C+]=N7)[C+]7[C+2][C+]"
                "(C8=C2[C+]=N8)[C+]2[C+](F)[C+2]C2([C+2]7)C2=C(N=[C+]2"
                ")[C+]2[C+2][C+]([C+2]C7([C+2][C+](F)[C+]27)C2=C6[C+]="
                "N2)C2=C(N=[C+]2)C2([C+2]4)[C+2][C+](F)[C+]52)[C+]13"
            ),
            name=name,
        ),
    ),
)
def cage_six_plus_nine(request) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )



from stk._internal.topology_graphs.rotaxane.nrotaxane import NRotaxane
from stk._internal.topology_graphs.rotaxane.vertices import (
    AxleVertex,
    CycleVertex,
)

__all__ = [
    "NRotaxane",
    "CycleVertex",
    "AxleVertex",
]



import pytest
import stk

from ..case_data import GenericCaseData


@pytest.fixture
def primary_amino(get_atom_ids):
    a, b, c, d = get_atom_ids(4)
    return _primary_amino(stk.N(a), stk.H(b), stk.H(c), stk.C(d))


def _primary_amino(nitrogen, hydrogen1, hydrogen2, atom):
    bonders = (nitrogen,)
    deleters = (hydrogen1, hydrogen2)
    return GenericCaseData(
        functional_group=stk.PrimaryAmino(
            nitrogen=nitrogen,
            hydrogen1=hydrogen1,
            hydrogen2=hydrogen2,
            atom=atom,
            bonders=bonders,
            deleters=deleters,
        ),
        atoms=(nitrogen, hydrogen1, hydrogen2, atom),
        bonders=bonders,
        deleters=deleters,
    )



from ..utilities import has_same_structure, is_clone


def test_clone(molecule):
    """
    Test :meth:`.Molecule.clone`.

    Parameters
    ----------
    molecule : :class:`.Molecule`
        The molecule to test.

    Returns
    -------
    None : :class:`NoneType`

    """

    clone = molecule.clone()
    has_same_structure(molecule, clone)
    is_clone(molecule, clone)



from collections import Counter

from ...utilities import DatabaseState, get_entry


def get_database_state(database):
    """
    Get the state of a :class:`.ValueMongoDb`.

    Parameters
    ----------
    database : :class:`.ValueMongoDb`
        The database whose state is wanted.

    Returns
    -------
    :class:`.DatabaseState`
        The current state of `database`.

    """

    entries = Counter()
    entries.update(map(get_entry, database._values.find({})))
    return DatabaseState(entries)



"""
Alcohol
=======

"""

from __future__ import annotations

import typing

from stk._internal.atom import Atom
from stk._internal.elements import H, O

from .generic_functional_group import GenericFunctionalGroup


class Alcohol(GenericFunctionalGroup):
    """
    Represents an alcohol functional group.

    The structure of the functional group is given by the pseudo-SMILES
    ``[atom][oxygen][hydrogen]``.

    """

    def __init__(
        self,
        oxygen: O,  # noqa: Not an ambiguous name.
        hydrogen: H,
        atom: Atom,
        bonders: tuple[Atom, ...],
        deleters: tuple[Atom, ...],
        placers: typing.Optional[tuple[Atom, ...]] = None,
    ) -> None:
        """
        Initialize a :class:`.Alcohol` instance.

        Parameters:

            oxygen:
                The oxygen atom.

            hydrogen:
                The hydrogen atom.

            atom:
                The atom to which the alcohol is attached.

            bonders:
                The bonder atoms.

            deleters:
                The deleter atoms.

            placers:
                The placer atoms. If ``None`` the `bonders` will be
                used.

        """

        self._oxygen = oxygen
        self._hydrogen = hydrogen
        self._atom = atom
        atoms = (oxygen, hydrogen, atom)
        super().__init__(
            atoms=atoms,
            bonders=bonders,
            deleters=deleters,
            placers=bonders if placers is None else placers,
        )

    def get_oxygen(self) -> O:  # noqa: Not an ambiguous name.
        """
        Get the oxygen atom.

        Returns:
            The oxygen atom.

        """

        return self._oxygen

    def get_hydrogen(self) -> H:
        """
        Get the hydrogen atom.

        Returns:
            The hydrogen atom.

        """

        return self._hydrogen

    def get_atom(self) -> Atom:
        """
        Get the atom to which the functional group is attached.

        Returns:
            The atom to which the functional group is attached.

        """

        return self._atom

    def clone(self) -> Alcohol:
        clone = self._clone()
        clone._oxygen = self._oxygen
        clone._hydrogen = self._hydrogen
        clone._atom = self._atom
        return clone

    def _with_ids(
        self,
        id_map: dict[int, int],
    ) -> Alcohol:
        super()._with_ids(id_map)

        if (oxygen_id := self._oxygen.get_id()) in id_map:
            self._oxygen = self._oxygen.with_id(id_map[oxygen_id])

        if (hydrogen_id := self._hydrogen.get_id()) in id_map:
            self._hydrogen = self._hydrogen.with_id(
                id=id_map[hydrogen_id],
            )

        if (atom_id := self._atom.get_id()) in id_map:
            self._atom = self._atom.with_id(id_map[atom_id])

        return self

    def with_ids(
        self,
        id_map: dict[int, int],
    ) -> Alcohol:
        return self.clone()._with_ids(id_map)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"{self._oxygen}, {self._hydrogen}, {self._atom}, "
            f"bonders={self._bonders}, deleters={self._deleters})"
        )



def test_get_num_building_block(case_data):
    """
    Test :meth:`.ConstructedMolecule.get_num_building_block`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the constructed molecule to test and the
        correct number of each building block.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_get_num_building_block(
        constructed_molecule=case_data.constructed_molecule,
        num_building_blocks=case_data.num_building_blocks,
    )


def _test_get_num_building_block(
    constructed_molecule,
    num_building_blocks,
):
    """
    Test :meth:`.ConstructedMolecule.get_num_building_blocks`.

    Parameters
    ----------
    constructed_molecule : :class:`.ConstructedMolecule`
        The constructed molecule to test.

    num_building_blocks : :class:`dict`
        Maps each building block in `constructed_molecule` to the
        number of times it was used during construction.

    Returns
    -------
    None : :class:`NoneType`

    """

    for building_block in constructed_molecule.get_building_blocks():
        assert (
            constructed_molecule.get_num_building_block(building_block)
            == num_building_blocks[building_block]
        )



import pytest
import stk

from ....case_data import CaseData


@pytest.fixture(
    scope="session",
    params=(
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.cage.EightPlusSixteen(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles="BrC1=C(Br)[C+]=N1",
                            functional_groups=[stk.BromoFactory()],
                        ),
                        stk.BuildingBlock(
                            smiles=(
                                "Br[C+]1C2(Br)[C+]=N[C+]2[C+](Br)[C+]("
                                "Br)[C+2]1"
                            ),
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                ),
            ),
            smiles=(
                "[C+]1=NC2=C1[C+]1[C+2][C+]3C4=C(N=[C+]4)[C+]4[C+2][C+"
                "]5C6=C([C+]=N6)[C+]6[C+2][C+]7C8=C([C+]=N8)C38[C+]=N["
                "C+]8[C+]1C1=C([C+]=N1)C13[C+]=N[C+]1[C+]1C8=C([C+]=N8"
                ")C89[C+]=N[C+]8[C+]8C%10=C([C+]=N%10)[C+]7[C+]7N=[C+]"
                "C67C6=C(N=[C+]6)C67[C+]=N[C+]6[C+]6C%10=C([C+]=N%10)["
                "C+]5[C+]5N=[C+]C45C4=C([C+]=N4)[C+]4[C+]2[C+2][C+]2C5"
                "=C([C+]=N5)[C+]3[C+2][C+]1C1=C(N=[C+]1)[C+]1[C+2][C+]"
                "(C3=C(N=[C+]3)C23[C+]=N[C+]43)[C+]2C3=C([C+]=N3)[C+]6"
                "[C+2][C+]7C3=C([C+]=N3)[C+]8[C+2][C+]9C3=C(N=[C+]3)C1"
                "3[C+]=N[C+]23"
            ),
            name=name,
        ),
    ),
)
def cage_eight_plus_sixteen(request) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )



from stk._internal.construction_state.construction_state import (
    ConstructionState,
)

from .utilities import _Placement


class _Serial:
    """
    Holds serial implementations of topology graph methods.
    """

    def __init__(self, stages: tuple[tuple[int, ...], ...]) -> None:
        """
        Parameters:
            stages:
                A :class:`tuple` of the form ``((v1, v2, v3), (v4, v5))``,
                where each nested :class:`tuple` holds the
                :class:`.Vertex` objects used for placement in a particular
                stage.
        """
        self._stages = stages

    def _place_building_blocks(
        self,
        state: ConstructionState,
    ) -> ConstructionState:
        for stage in self._stages:
            vertices = tuple(state.get_vertices(stage))
            building_blocks = tuple(map(state.get_building_block, stage))
            edges = tuple(map(state.get_edges, stage))
            placements = map(
                _Placement,
                vertices,
                edges,
                building_blocks,
            )
            placement_results = map(
                lambda placement: placement.get_result(),
                placements,
            )
            state = state.with_placement_results(
                vertices=vertices,
                edges=edges,
                building_blocks=building_blocks,
                results=placement_results,
            )
        return state

    def get_num_stages(self) -> int:
        """
        Get the number of placement stages.

        Returns:
            The number of placement stages.
        """
        return len(self._stages)



from .case_data import CaseData


def test_get_num_functional_groups(case_data: CaseData) -> None:
    """
    Test :meth:`.BuildingBlock.get_num_functional_groups`.

    Parameters:
        case_data:
            A test case. Holds the building block to test and the correct
            functional groups.

    """

    assert case_data.building_block.get_num_functional_groups() == len(
        case_data.functional_groups
    )



import pytest
from pytest_lazyfixture import lazy_fixture

from .case_data import CaseData

# Fixtures need to be visible for lazy_fixture() calls.
from .fixtures import *  # noqa


@pytest.fixture(
    params=(lazy_fixture("constructed_molecule_mongo_db"),),
)
def case_data(request) -> CaseData:
    return request.param



import pytest
import stk


def get_atom_map_0(functional_group):
    """
    Get an atom_map with a single atom.

    """

    # Make sure new_id is always valid, by making 1 larger than the
    # biggest on in the functional group. This prevents two atoms in
    # the functional group from having the same id.
    new_id = max(functional_group.get_atom_ids()) + 1
    atoms = (stk.Li(new_id),)
    return dict(zip(functional_group.get_atom_ids(), atoms))


@pytest.fixture(
    params=[
        lambda functional_group: {},
        get_atom_map_0,
    ],
)
def get_atom_map(request):
    """
    Return a valid `atom_map` parameter for a functional group.

    Parameters
    ----------
    functional_group : :class:`.FunctionalGroup`
        A functional group being cloned, for which a valid
        `atom_map` parameter needs to created.

    Returns
    -------
    :class:`dict`
        A valid `atom_map` parameter for calling
        :meth:`~.FunctionalGroup.with_atoms` on `functional_group`.

    """

    return request.param



import itertools
import typing
from collections.abc import Callable, Iterable, Iterator, Sequence

from stk._internal.ea.molecule_record import MoleculeRecord
from stk._internal.ea.selection.batch import Batch
from stk._internal.ea.selection.selectors.yielded_batches import YieldedBatches
from stk._internal.key_makers.inchi import Inchi
from stk._internal.key_makers.molecule import MoleculeKeyMaker

from .selector import Selector

T = typing.TypeVar("T", bound=MoleculeRecord)


class Worst(Selector[T]):
    """
    Selects batches of molecules, lowest fitness value first.

    Examples:

        *Yielding Batches Holding Multiple Molecules*

        Select the worst 5 batches of size 3

        .. testcode:: yielding-batches-holding-multiple-molecules

            import stk

            worst = stk.Worst(num_batches=5, batch_size=3)
            population = {
                stk.MoleculeRecord(
                    topology_graph=stk.polymer.Linear(
                        building_blocks=[
                            stk.BuildingBlock('BrCCBr', stk.BromoFactory()),
                        ],
                        repeating_unit='A',
                        num_repeating_units=2,
                    ),
                ): i
                for i in range(10)
            }
            for batch in worst.select(population):
                # Do stuff with batch.
                pass
    """

    def __init__(
        self,
        num_batches: int | None = None,
        batch_size: int = 1,
        duplicate_molecules: bool = True,
        duplicate_batches: bool = True,
        key_maker: MoleculeKeyMaker = Inchi(),
        fitness_modifier: Callable[
            [dict[T, float]], dict[T, float]
        ] = lambda x: x,
    ) -> None:
        """
        Parameters:
            num_batches:
                The number of batches to yield. If ``None`` then yielding
                will continue forever or until the generator is exhausted,
                whichever comes first.

            batch_size:
                The number of molecules yielded at once.

            duplicate_molecules:
                If ``True`` the same molecule can be yielded in more than
                one batch.

            duplicate_batches:
                If ``True`` the same batch can be yielded more than once.
                Duplicate batches can occur if the same molecule is found
                multiple times in a population.

            key_maker:
                Used to get the keys of molecules, which are used to
                determine if two molecules are duplicates of each
                other.

            fitness_modifier:
                Takes the `population` on which :meth:`~.Selector.select`
                is called and returns a :class:`dict`, which maps records
                in the `population` to the fitness values the
                :class:`.Selector` should use. If ``None``, the regular
                fitness values of the records are used.
        """
        super().__init__(key_maker, fitness_modifier, batch_size)

        self._duplicate_molecules = duplicate_molecules
        self._duplicate_batches = duplicate_batches
        self._num_batches = num_batches

    def _select_from_batches(
        self,
        batches: Sequence[Batch[T]],
        yielded_batches: YieldedBatches[T],
    ) -> Iterator[Batch[T]]:
        selected_batches: Iterable[Batch[T]] = sorted(batches)

        if not self._duplicate_molecules:
            selected_batches = filter(
                yielded_batches.has_no_yielded_molecules,
                selected_batches,
            )

        if not self._duplicate_batches:
            selected_batches = filter(
                yielded_batches.is_unyielded_batch,
                selected_batches,
            )

        yield from itertools.islice(selected_batches, self._num_batches)



class CaseData:
    """
    A test case.

    Attributes
    ----------
    database : class:`.ValueDatabase`
        The database to test.

    molecule : :class:`.Molecule`
        The molecule to test.

    value : :class:`object`
        The value to put into the database.

    """

    def __init__(self, database, molecule, value):
        """
        Initialize a :class:`.CaseData` instance.

        Parameters
        ----------
        database : class:`.ValueDatabase`
            The database to test.

        molecule : :class:`.Molecule`
            The molecule to test.

        value : :class:`object`
            The value to put into the database.

        """

        self.database = database
        self.molecule = molecule
        self.value = value



import typing
from collections.abc import Sequence
from dataclasses import dataclass

import stk

T = typing.TypeVar("T", bound=stk.MoleculeRecord)


@dataclass(frozen=True, slots=True)
class CaseData(typing.Generic[T]):
    selector: stk.Selector[T]
    population: dict[T, float]
    selected: Sequence[stk.Batch[T]]

    @staticmethod
    def new(
        selector: stk.Selector[T],
        population: dict[T, float],
        selected: Sequence[Sequence[int]],
    ) -> "CaseData":
        records = list(population)
        batches = []
        for batch in selected:
            batches.append(
                stk.Batch(
                    records={
                        records[member]: population[records[member]]
                        for member in batch
                    },
                    key_maker=stk.Inchi(),
                )
            )
        return CaseData(selector, population, batches)
